# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger('pyomo.network')

from pyomo.common.log import is_debug_set
from pyomo.common.modeling import unique_component_name

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.core.base.indexed_component import UnindexedComponent_set
from pyomo.core.base import Transformation, Block, SortComponents, TransformationFactory
from pyomo.gdp import Disjunct

from pyomo.network import Arc
from pyomo.network.util import replicate_var

# keyword arguments for component_objects and component_data_objects
obj_iter_kwds = dict(
    ctype=Arc,
    active=True,
    sort=SortComponents.deterministic,
    descend_into=(Block, Disjunct),
)


@TransformationFactory.register(
    'network.expand_arcs', doc="Expand all Arcs in the model to simple constraints"
)
class ExpandArcs(Transformation):
    def _apply_to(self, instance, **kwds):
        if is_debug_set(logger):
            logger.debug("Calling ArcExpander")

        # need to collect all ports to see every port each
        # is related to so that we can expand empty ports
        port_list, known_port_sets, matched_ports = self._collect_ports(instance)

        self._add_blocks(instance)

        for port in port_list:
            # iterate over ref so that the index set is the same
            # for all occurrences of this member in related ports
            # and so we iterate over members deterministically
            ref = known_port_sets[id(matched_ports[port])]
            for k, v in sorted(ref.items()):
                rule, kwds = port._rules[k]
                if v[1] >= 0:
                    index_set = v[0].index_set()
                else:
                    index_set = UnindexedComponent_set
                rule(port, k, index_set, **kwds)

        for arc in instance.component_objects(**obj_iter_kwds):
            arc.deactivate()

    def _collect_ports(self, instance):
        # List of the ports in the order in which we found them
        # (this should be deterministic, provided that the user's model
        # is deterministic)
        port_list = []
        # ID of the next port group (set of matched ports)
        groupID = 0
        # port_groups stars out as a dict of {id(set): (groupID, set)}
        # If you sort by the groupID, then this will be deterministic.
        port_groups = dict()
        # map of port to the set of ports that must match it
        matched_ports = ComponentMap()

        for arc in instance.component_data_objects(**obj_iter_kwds):
            ref = None

            for p in arc.ports:
                if p in matched_ports:
                    if ref is None:
                        # The first port in this arc has
                        # already been seen. We will use that Set as
                        # the reference
                        ref = matched_ports[p]
                    elif ref is not matched_ports[p]:
                        # We already have a reference group; merge this
                        # new group into it.

                        # Optimization: this merge is linear in the size
                        # of the src set. If the reference set is
                        # smaller, save time by switching to a new
                        # reference set.
                        src = matched_ports[p]
                        if len(ref) < len(src):
                            ref, src = src, ref
                        ref.update(src)
                        for i in src:
                            matched_ports[i] = ref
                        del port_groups[id(src)]
                    # else: pass
                    #   The new group *is* the reference group;
                    #   there is nothing to do.
                else:
                    # The port has not been seen before.
                    port_list.append(p)
                    if ref is None:
                        # This is the first port in the arc:
                        # start a new reference set.
                        ref = ComponentSet()
                        port_groups[id(ref)] = (groupID, ref)
                        groupID += 1
                    # This port hasn't been seen. Record it.
                    ref.add(p)
                    matched_ports[p] = ref

        # Validate all port sets and expand the empty ones
        known_port_sets = {}
        for groupID, port_set in sorted(port_groups.values()):
            known_port_sets[id(port_set)] = self._validate_and_expand_port_set(port_set)

        return port_list, known_port_sets, matched_ports

    def _validate_and_expand_port_set(self, ports):
        ref = {}
        # First, go through the ports and get the superset of all fields
        for p in ports:
            for k, v in p.vars.items():
                if k in ref:
                    # We have already seen this var
                    continue
                if v is None:
                    # This is an implicit var
                    continue
                # OK: New var, so add it to the reference list
                _len = -1 if not v.is_indexed() else len(v)
                ref[k] = (v, _len, p, p.rule_for(k))

        if not ref:
            logger.warning(
                "Cannot identify a reference port: no ports "
                "in the port set have assigned variables:\n\t(%s)"
                % ', '.join(sorted(p.name for p in ports.values()))
            )
            return ref

        # Now make sure that ports match
        empty_or_partial = []
        for p in ports:
            p_is_partial = False
            if not p.vars:
                # This is an empty port and should be defined with
                # "auto" vars
                empty_or_partial.append(p)
                continue

            for k, v in ref.items():
                if k not in p.vars:
                    raise ValueError(
                        "Port mismatch: Port '%s' missing variable "
                        "'%s' (appearing in reference port '%s')"
                        % (p.name, k, v[2].name)
                    )
                _v = p.vars[k]
                if _v is None:
                    if not p_is_partial:
                        empty_or_partial.append(p)
                        p_is_partial = True
                    continue
                _len = -1 if not _v.is_indexed() else len(_v)
                if (_len >= 0) ^ (v[1] >= 0):
                    raise ValueError(
                        "Port mismatch: Port variable '%s' mixing "
                        "indexed and non-indexed targets on ports '%s' "
                        "and '%s'" % (k, v[2].name, p.name)
                    )
                if _len >= 0 and _len != v[1]:
                    raise ValueError(
                        "Port mismatch: Port variable '%s' index "
                        "mismatch (%s elements in reference port '%s', "
                        "but %s elements in port '%s')"
                        % (k, v[1], v[2].name, _len, p.name)
                    )
                if v[1] >= 0 and len(v[0].index_set() ^ _v.index_set()):
                    raise ValueError(
                        "Port mismatch: Port variable '%s' has "
                        "mismatched indices on ports '%s' and '%s'"
                        % (k, v[2].name, p.name)
                    )
                if p.rule_for(k) is not v[3]:
                    raise ValueError(
                        "Port mismatch: Port variable '%s' has "
                        "different rules on ports '%s' and '%s'"
                        % (k, v[2].name, p.name)
                    )

        # as we are adding things to the model, sort by key so that
        # the order things are added is deterministic
        sorted_refs = sorted(ref.items())
        if len(empty_or_partial) > 1:
            # This is expensive (names aren't cheap), but does result in
            # a deterministic ordering
            empty_or_partial.sort(key=lambda x: x.getname(fully_qualified=True))

        # Fill in any empty ports
        for p in empty_or_partial:
            block = p.parent_block()
            for k, v in sorted_refs:
                if k in p.vars and p.vars[k] is not None:
                    continue

                vname = unique_component_name(
                    block, '%s_auto_%s' % (p.getname(fully_qualified=True), k)
                )

                new_var = replicate_var(v[0], vname, block)

                # add this new variable to the port so that it has a rule
                p.add(new_var, k, rule=v[3])

        return ref

    def _add_blocks(self, instance):
        # iterate over component_objects so we can make indexed blocks
        for arc in instance.component_objects(**obj_iter_kwds):
            blk = Block(arc.index_set())
            bname = unique_component_name(
                arc.parent_block(), "%s_expanded" % arc.local_name
            )
            arc.parent_block().add_component(bname, blk)
            arc._expanded_block = blk
            if arc.is_indexed():
                for i in arc:
                    arc[i]._expanded_block = blk[i]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.kernel as pmo
from pyomo.core.kernel.block import IBlock
from pyomo.core.kernel.variable import IVariable
from pyomo.core.kernel.constraint import IConstraint


class IJunk(IBlock):
    __slots__ = ()


class junk(pmo.block):
    _ctype = IJunk


class junk_list(pmo.block_list):
    __slots__ = ()
    _ctype = IJunk


_model = pmo.block()
_model.v = pmo.variable()
_model.V = pmo.variable_list()
_model.V.append(pmo.variable())
_model.V.append(pmo.variable_list())
_model.V[1].append(pmo.variable())
_model.c = pmo.constraint()
_model.C = pmo.constraint_list()
_model.C.append(pmo.constraint())
_model.C.append(pmo.constraint_list())
_model.C[1].append(pmo.constraint())
b_clone = _model.clone()
_model.b = b_clone.clone()
_model.B = pmo.block_list()
_model.B.append(b_clone.clone())
_model.B.append(pmo.block_list())
_model.B[1].append(b_clone.clone())
del b_clone
_model.j = junk()
_model.J = junk_list()
_model.J.append(junk())
_model.J.append(junk_list())
_model.J[1].append(junk())
_model.J[1][0].b = pmo.block()
_model.J[1][0].b.v = pmo.variable()
model_clone = _model.clone()
_model.k = pmo.block()
_model.K = pmo.block_list()
_model.K.append(model_clone.clone())
del model_clone


class Test_kernel(unittest.TestCase):
    def test_no_ctype_collisions(self):
        hash_set = set()
        hash_list = list()
        for cls in [
            pmo.variable,
            pmo.constraint,
            pmo.objective,
            pmo.expression,
            pmo.parameter,
            pmo.suffix,
            pmo.sos,
            pmo.block,
        ]:
            ctype = cls._ctype
            hash_set.add(hash(ctype))
            hash_list.append(hash(ctype))
        self.assertEqual(len(hash_set), len(hash_list))

    def test_component_data_objects_hack(self):
        model = _model.clone()
        self.assertEqual(
            [str(obj) for obj in model.component_data_objects()],
            [str(obj) for obj in model.components()],
        )
        self.assertEqual(
            [str(obj) for obj in model.component_data_objects(ctype=IVariable)],
            [str(obj) for obj in model.components(ctype=IVariable)],
        )
        self.assertEqual(
            [str(obj) for obj in model.component_data_objects(ctype=IConstraint)],
            [str(obj) for obj in model.components(ctype=IConstraint)],
        )
        self.assertEqual(
            [str(obj) for obj in model.component_data_objects(ctype=IBlock)],
            [str(obj) for obj in model.components(ctype=IBlock)],
        )
        self.assertEqual(
            [str(obj) for obj in model.component_data_objects(ctype=IJunk)],
            [str(obj) for obj in model.components(ctype=IJunk)],
        )
        for item in pmo.preorder_traversal(model):
            item.deactivate()
            self.assertEqual(
                [str(obj) for obj in model.component_data_objects(active=True)],
                [str(obj) for obj in model.components(active=True)],
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in model.component_data_objects(
                        ctype=IVariable, active=True
                    )
                ],
                [str(obj) for obj in model.components(ctype=IVariable, active=True)],
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in model.component_data_objects(
                        ctype=IConstraint, active=True
                    )
                ],
                [str(obj) for obj in model.components(ctype=IConstraint, active=True)],
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in model.component_data_objects(ctype=IBlock, active=True)
                ],
                [str(obj) for obj in model.components(ctype=IBlock, active=True)],
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in model.component_data_objects(ctype=IJunk, active=True)
                ],
                [str(obj) for obj in model.components(ctype=IJunk, active=True)],
            )
            item.activate()

    def test_component_objects_hack(self):
        model = _model.clone()
        objs = {key: [] for key in [None, IVariable, IConstraint, IBlock, IJunk]}
        for item in pmo.heterogeneous_containers(model):
            objs[None].extend(item.component_objects(descend_into=False))
            self.assertEqual(
                [str(obj) for obj in item.component_objects(descend_into=False)],
                [str(obj) for obj in item.children()],
            )
            objs[IVariable].extend(
                item.component_objects(ctype=IVariable, descend_into=False)
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in item.component_objects(
                        ctype=IVariable, descend_into=False
                    )
                ],
                [str(obj) for obj in item.children(ctype=IVariable)],
            )
            objs[IConstraint].extend(
                item.component_objects(ctype=IConstraint, descend_into=False)
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in item.component_objects(
                        ctype=IConstraint, descend_into=False
                    )
                ],
                [str(obj) for obj in item.children(ctype=IConstraint)],
            )
            objs[IBlock].extend(
                item.component_objects(ctype=IBlock, descend_into=False)
            )
            self.assertEqual(
                [
                    str(obj)
                    for obj in item.component_objects(ctype=IBlock, descend_into=False)
                ],
                [str(obj) for obj in item.children(ctype=IBlock)],
            )
            objs[IJunk].extend(item.component_objects(ctype=IJunk, descend_into=False))
            self.assertEqual(
                [
                    str(obj)
                    for obj in item.component_objects(ctype=IJunk, descend_into=False)
                ],
                [str(obj) for obj in item.children(ctype=IJunk)],
            )
        all_ = []
        for key in objs:
            if key is None:
                continue
            names = [str(obj) for obj in objs[key]]
            self.assertEqual(
                sorted([str(obj) for obj in model.component_objects(ctype=key)]),
                sorted(names),
            )
            all_.extend(names)
        self.assertEqual(
            sorted([str(obj) for obj in model.component_objects()]), sorted(all_)
        )
        self.assertEqual(sorted([str(obj) for obj in objs[None]]), sorted(all_))
        model.deactivate()
        self.assertEqual(
            sorted([str(obj) for obj in model.component_objects()]), sorted(all_)
        )
        self.assertEqual(
            [
                str(obj)
                for obj in model.component_objects(descend_into=False, active=True)
            ],
            [],
        )
        self.assertEqual(
            [
                str(obj)
                for obj in model.component_objects(descend_into=True, active=True)
            ],
            [],
        )

    def test_block_data_objects_hack(self):
        model = _model.clone()
        model.deactivate()
        self.assertEqual(
            [str(obj) for obj in model.block_data_objects(active=True)], []
        )
        self.assertEqual(
            [str(obj) for obj in model.block_data_objects()],
            [str(model)] + [str(obj) for obj in model.components(ctype=IBlock)],
        )
        model.activate()
        self.assertEqual(
            [str(obj) for obj in model.block_data_objects(active=True)],
            [str(model)] + [str(obj) for obj in model.components(ctype=IBlock)],
        )
        self.assertEqual(
            [str(obj) for obj in model.block_data_objects()],
            [str(model)] + [str(obj) for obj in model.components(ctype=IBlock)],
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import time
import logging
import array
from weakref import ref as weakref_ref

from pyomo.common.log import is_debug_set
from pyomo.common.numeric_types import value
from pyomo.common.modeling import NOTSET
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import is_fixed, ZeroConstant
from pyomo.core.base.set_types import Any
from pyomo.core.base import SortComponents, Var
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.constraint import (
    Constraint,
    IndexedConstraint,
    ScalarConstraint,
    ConstraintData,
)
from pyomo.core.expr.numvalue import native_numeric_types
from pyomo.repn import generate_standard_repn

from collections.abc import Mapping

logger = logging.getLogger('pyomo.core')


def _label_bytes(x):
    if x < 1e3:
        return str(x) + " B"
    if x < 1e6:
        return str(x / 1.0e3) + " KB"
    if x < 1e9:
        return str(x / 1.0e6) + " MB"
    return str(x / 1.0e9) + " GB"


#
# Compile a Pyomo constructed model in-place, storing the compiled
# sparse constraint object on the model under constraint_name.
#
def compile_block_linear_constraints(
    parent_block,
    constraint_name,
    skip_trivial_constraints=False,
    single_precision_storage=False,
    verbose=False,
    descend_into=True,
):
    if verbose:
        print("")
        print(
            "Compiling linear constraints on block with name: %s" % (parent_block.name)
        )

    if not parent_block.is_constructed():
        raise RuntimeError(
            "Attempting to compile block '%s' with unconstructed "
            "component(s)" % (parent_block.name)
        )

    #
    # Linear MatrixConstraint in CSR format
    #
    SparseMat_pRows = []
    SparseMat_jCols = []
    SparseMat_Vals = []
    Ranges = []
    RangeTypes = []

    def _get_bound(exp):
        if exp is None:
            return None
        if is_fixed(exp):
            return value(exp)
        raise ValueError("non-fixed bound: " + str(exp))

    start_time = time.time()
    if verbose:
        print("Sorting active blocks...")

    sortOrder = SortComponents.indices | SortComponents.alphabetical
    all_blocks = [
        _b
        for _b in parent_block.block_data_objects(
            active=True, sort=sortOrder, descend_into=descend_into
        )
    ]

    stop_time = time.time()
    if verbose:
        print("Time to sort active blocks: %.2f seconds" % (stop_time - start_time))

    start_time = time.time()
    if verbose:
        print("Collecting variables on active blocks...")

    #
    # First Pass: assign each variable a deterministic id
    #             (an index in a list)
    #
    VarSymbolToVarObject = []
    for block in all_blocks:
        VarSymbolToVarObject.extend(
            block.component_data_objects(Var, sort=sortOrder, descend_into=False)
        )
    VarIDToVarSymbol = dict(
        (id(vardata), index) for index, vardata in enumerate(VarSymbolToVarObject)
    )

    stop_time = time.time()
    if verbose:
        print(
            "Time to collect variables on active blocks: %.2f seconds"
            % (stop_time - start_time)
        )

    start_time = time.time()
    if verbose:
        print("Compiling active linear constraints...")

    #
    # Second Pass: collect and remove active linear constraints
    #
    constraint_data_to_remove = []
    empty_constraint_containers_to_remove = []
    constraint_containers_to_remove = []
    constraint_containers_to_check = set()
    referenced_variable_symbols = set()
    nnz = 0
    nrows = 0
    SparseMat_pRows = [0]
    for block in all_blocks:
        if hasattr(block, '_repn'):
            del block._repn

        for constraint in block.component_objects(
            Constraint, active=True, sort=sortOrder, descend_into=False
        ):
            assert not isinstance(constraint, MatrixConstraint)

            if len(constraint) == 0:
                empty_constraint_containers_to_remove.append((block, constraint))

            else:
                singleton = isinstance(constraint, ScalarConstraint)

                # Note that as we may be removing items from the _data
                # dictionary, we need to make a copy of the items list
                # before iterating:
                for index, constraint_data in list(constraint.items()):
                    if (
                        constraint_data.body.__class__ in native_numeric_types
                        or constraint_data.body.polynomial_degree() <= 1
                    ):
                        # collect for removal
                        if singleton:
                            constraint_containers_to_remove.append((block, constraint))
                        else:
                            constraint_data_to_remove.append((constraint, index))
                            constraint_containers_to_check.add((block, constraint))

                        repn = generate_standard_repn(constraint_data.body)

                        assert repn.nonlinear_expr is None

                        row_variable_symbols = []
                        row_coefficients = []
                        if len(repn.linear_vars) == 0:
                            if skip_trivial_constraints:
                                continue
                        else:
                            row_variable_symbols = [
                                VarIDToVarSymbol[id(vardata)]
                                for vardata in repn.linear_vars
                            ]
                            referenced_variable_symbols.update(row_variable_symbols)
                            assert repn.linear_coefs is not None
                            row_coefficients = repn.linear_coefs

                        SparseMat_pRows.append(
                            SparseMat_pRows[-1] + len(row_variable_symbols)
                        )
                        SparseMat_jCols.extend(row_variable_symbols)
                        SparseMat_Vals.extend(row_coefficients)

                        nnz += len(row_variable_symbols)
                        nrows += 1

                        L = _get_bound(constraint_data.lower)
                        U = _get_bound(constraint_data.upper)
                        constant = value(repn.constant)

                        Ranges.append(L - constant if (L is not None) else 0)
                        Ranges.append(U - constant if (U is not None) else 0)
                        if (
                            (L is not None)
                            and (U is not None)
                            and (not constraint_data.equality)
                        ):
                            RangeTypes.append(
                                MatrixConstraint.LowerBound
                                | MatrixConstraint.UpperBound
                            )
                        elif constraint_data.equality:
                            RangeTypes.append(MatrixConstraint.Equality)
                        elif L is not None:
                            assert U is None
                            RangeTypes.append(MatrixConstraint.LowerBound)
                        else:
                            assert U is not None
                            RangeTypes.append(MatrixConstraint.UpperBound)

                        # Start freeing up memory
                        constraint[index] = Constraint.Skip

    ncols = len(referenced_variable_symbols)

    stop_time = time.time()
    if verbose:
        print(
            "Time to compile active linear constraints: %.2f seconds"
            % (stop_time - start_time)
        )

    start_time = time.time()
    if verbose:
        print("Removing compiled constraint objects...")

    #
    # Remove compiled constraints
    #
    constraints_removed = 0
    constraint_containers_removed = 0
    for block, constraint in empty_constraint_containers_to_remove:
        block.del_component(constraint)
        constraint_containers_removed += 1
    for constraint, index in constraint_data_to_remove:
        # Note that this del is not needed: assigning Constraint.Skip
        # above removes the ConstraintData from the _data dict.
        # del constraint[index]
        constraints_removed += 1
    for block, constraint in constraint_containers_to_remove:
        block.del_component(constraint)
        constraints_removed += 1
        constraint_containers_removed += 1
    for block, constraint in constraint_containers_to_check:
        if len(constraint) == 0:
            block.del_component(constraint)
            constraint_containers_removed += 1

    stop_time = time.time()
    if verbose:
        print(
            "Eliminated %s constraints and %s Constraint container objects"
            % (constraints_removed, constraint_containers_removed)
        )
        print(
            "Time to remove compiled constraint objects: %.2f seconds"
            % (stop_time - start_time)
        )

    start_time = time.time()
    if verbose:
        print("Assigning variable column indices...")

    #
    # Assign a column index to the set of referenced variables
    #
    ColumnIndexToVarSymbol = sorted(referenced_variable_symbols)
    VarSymbolToColumnIndex = dict(
        (symbol, column) for column, symbol in enumerate(ColumnIndexToVarSymbol)
    )
    SparseMat_jCols = [VarSymbolToColumnIndex[symbol] for symbol in SparseMat_jCols]
    del VarSymbolToColumnIndex
    ColumnIndexToVarObject = [
        VarSymbolToVarObject[var_symbol] for var_symbol in ColumnIndexToVarSymbol
    ]

    stop_time = time.time()
    if verbose:
        print(
            "Time to assign variable column indices: %.2f seconds"
            % (stop_time - start_time)
        )

    start_time = time.time()
    if verbose:
        print("Converting compiled constraint data to array storage...")
        print(
            "  - Using %s precision for numeric values"
            % ('single' if single_precision_storage else 'double')
        )

    #
    # Convert to array storage
    #

    number_storage = 'f' if single_precision_storage else 'd'
    SparseMat_pRows = array.array('L', SparseMat_pRows)
    SparseMat_jCols = array.array('L', SparseMat_jCols)
    SparseMat_Vals = array.array(number_storage, SparseMat_Vals)
    Ranges = array.array(number_storage, Ranges)
    RangeTypes = array.array('B', RangeTypes)

    stop_time = time.time()
    if verbose:
        storage_bytes = (
            SparseMat_pRows.buffer_info()[1] * SparseMat_pRows.itemsize
            + SparseMat_jCols.buffer_info()[1] * SparseMat_jCols.itemsize
            + SparseMat_Vals.buffer_info()[1] * SparseMat_Vals.itemsize
            + Ranges.buffer_info()[1] * Ranges.itemsize
            + RangeTypes.buffer_info()[1] * RangeTypes.itemsize
        )
        print("Sparse Matrix Dimension:")
        print("  - Rows: " + str(nrows))
        print("  - Cols: " + str(ncols))
        print("  - Nonzeros: " + str(nnz))
        print("Compiled Data Storage: " + str(_label_bytes(storage_bytes)))
        print(
            "Time to convert compiled constraint data to "
            "array storage: %.2f seconds" % (stop_time - start_time)
        )

    parent_block.add_component(
        constraint_name,
        MatrixConstraint(
            nrows,
            ncols,
            nnz,
            SparseMat_pRows,
            SparseMat_jCols,
            SparseMat_Vals,
            Ranges,
            RangeTypes,
            ColumnIndexToVarObject,
        ),
    )


# class _LinearConstraintData(ConstraintData,LinearCanonicalRepn):
#
# This change breaks this class, but it's unclear whether this
# is being used...
#
class _LinearConstraintData(ConstraintData):
    """
    This class defines the data for a single linear constraint
        in canonical form.

    Constructor arguments:
        component       The Constraint object that owns this data.

    Public class attributes:
        active          A boolean that is true if this constraint is
                            active in the model.
        body            The Pyomo expression for this constraint
        lower           The Pyomo expression for the lower bound
        upper           The Pyomo expression for the upper bound
        equality        A boolean that indicates whether this is an
                            equality constraint
        strict_lower    A boolean that indicates whether this
                            constraint uses a strict lower bound
        strict_upper    A boolean that indicates whether this
                            constraint uses a strict upper bound
        variables       A tuple of variables comprising the body
                            of this constraint
        coefficients    A tuple of coefficients matching the order
                            of variables that comprise the body of
                            this constraint
        constant        A number representing the aggregation of any
                            constant/fixed items found in the body of
                            this constraint

    Private class attributes:
        _component      The objective component.
        _active         A boolean that indicates whether this data is active
    """

    __slots__ = ()

    def __init__(self, index, component=None):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - ConstraintData,
        #   - ActiveComponentData
        #   - ComponentData
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = index
        self._active = True


class _LinearMatrixConstraintData(_LinearConstraintData):
    """
    This class defines the data for a single linear constraint
        derived from a canonical form Ax=b constraint.

    Constructor arguments:
        component       The Constraint object that owns this data.

    Public class attributes:
        active          A boolean that is true if this constraint is
                            active in the model.
        body            The Pyomo expression for this constraint
        lower           The Pyomo expression for the lower bound
        upper           The Pyomo expression for the upper bound
        equality        A boolean that indicates whether this is an
                            equality constraint
        strict_lower    A boolean that indicates whether this
                            constraint uses a strict lower bound
        strict_upper    A boolean that indicates whether this
                            constraint uses a strict upper bound
        variables       A tuple of variables comprising the body
                            of this constraint
        coefficients    A tuple of coefficients matching the order
                            of variables that comprise the body of
                            this constraint
        constant        A number representing the aggregate of any
                            constants found in the body of this
                            constraint

    Private class attributes:
        _component      The objective component.
        _active         A boolean that indicates whether this data is active
    """

    __slots__ = ()

    def __init__(self, index, component=None):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - _LinearConstraintData
        #   - ConstraintData,
        #   - ActiveComponentData
        #   - ComponentData
        self._component = weakref_ref(component) if (component is not None) else None
        self._active = True

        # row index into the sparse matrix stored on the parent
        assert index >= 0
        self._index = index

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        result = super(_LinearMatrixConstraintData, self).__getstate__()
        return result

    # Since this class requires no special processing of the state
    # dictionary, it does not need to implement __setstate__()

    #
    # Override the default interface methods to
    # avoid generating the body expression where
    # possible
    #

    def __call__(self, exception=NOTSET):
        """
        Compute the value of the body of this constraint.
        """
        exception = _type_check_exception_arg(self, exception)
        comp = self.parent_component()
        index = self.index()
        prows = comp._prows
        jcols = comp._jcols
        varmap = comp._varmap
        vals = comp._vals
        try:
            return sum(
                varmap[jcols[p]]() * vals[p]
                for p in range(prows[index], prows[index + 1])
            )
        except (ValueError, TypeError):
            if exception:
                raise
            return None

    def has_lb(self):
        """Returns :const:`False` when the lower bound is
        :const:`None` or negative infinity"""
        lb = self.lower
        return (lb is not None) and (lb != float('-inf'))

    def has_ub(self):
        """Returns :const:`False` when the upper bound is
        :const:`None` or positive infinity"""
        ub = self.upper
        return (ub is not None) and (ub != float('inf'))

    def lslack(self):
        """
        Returns the value of L-f(x) for constraints of the form:
            L <= f(x) (<= U)
            (U >=) f(x) >= L
        """
        raise self.lower - self()

    def uslack(self):
        """
        Returns the value of U-f(x) for constraints of the form:
            (L <=) f(x) <= U
            U >= f(x) (>= L)
        """
        self.upper - self()

    #
    # Override some default implementations on ComponentData
    #

    def index(self):
        return self._index

    #
    # Abstract Interface (LinearCanonicalRepn)
    #

    @property
    def variables(self):
        """A tuple of variables comprising the constraint body."""
        comp = self.parent_component()
        prows = comp._prows
        jcols = comp._jcols
        varmap = comp._varmap
        if prows[self._index] == prows[self._index + 1]:
            return ()
        variables = tuple(
            varmap[jcols[p]]
            for p in range(prows[self._index], prows[self._index + 1])
            if not varmap[jcols[p]].fixed
        )

        return variables

    @property
    def coefficients(self):
        """A tuple of coefficients associated with the variables."""
        comp = self.parent_component()
        prows = comp._prows
        jcols = comp._jcols
        vals = comp._vals
        varmap = comp._varmap
        if prows[self._index] == prows[self._index + 1]:
            return ()
        coefs = tuple(
            vals[p]
            for p in range(prows[self._index], prows[self._index + 1])
            if not varmap[jcols[p]].fixed
        )

        return coefs

    # for backwards compatibility
    linear = coefficients

    @property
    def constant(self):
        """The constant value associated with the constraint body."""
        comp = self.parent_component()
        prows = comp._prows
        jcols = comp._jcols
        vals = comp._vals
        varmap = comp._varmap
        if prows[self._index] == prows[self._index + 1]:
            return 0
        terms = tuple(
            vals[p] * varmap[jcols[p]]()
            for p in range(prows[self._index], prows[self._index + 1])
            if varmap[jcols[p]].fixed
        )

        return sum(terms)

    #
    # Abstract Interface (ConstraintData)
    #

    def to_bounded_expression(self, evaluate_bounds=False):
        """Access this constraint as a single expression."""
        # Note that the bounds are always going to be floats...
        return self.lower, self.body, self.upper

    @property
    def body(self):
        """Access the body of a constraint expression."""
        comp = self.parent_component()
        index = self.index()
        prows = comp._prows
        jcols = comp._jcols
        varmap = comp._varmap
        vals = comp._vals
        if prows[self._index] == prows[self._index + 1]:
            return ZeroConstant
        return sum(
            varmap[jcols[p]] * vals[p] for p in range(prows[index], prows[index + 1])
        )

    @property
    def lower(self):
        """Access the lower bound of a constraint expression."""
        comp = self.parent_component()
        index = self.index()
        if comp._range_types[index] & MatrixConstraint.LowerBound:
            return comp._ranges[2 * index]
        return None

    @property
    def upper(self):
        """Access the upper bound of a constraint expression."""
        comp = self.parent_component()
        index = self.index()
        if comp._range_types[index] & MatrixConstraint.UpperBound:
            return comp._ranges[(2 * index) + 1]
        return None

    @property
    def lb(self):
        """Access the lower bound of a constraint expression."""
        return self.lower

    @property
    def ub(self):
        """Access the upper bound of a constraint expression."""
        return self.upper

    @property
    def equality(self):
        """A boolean indicating whether this is an equality constraint."""
        return (
            self.parent_component()._range_types[self.index()]
            & MatrixConstraint.Equality
        ) == MatrixConstraint.Equality

    @property
    def strict_lower(self):
        """A boolean indicating whether this constraint has a strict lower bound."""
        return (
            self.parent_component()._range_types[self.index()]
            & MatrixConstraint.StrictLowerBound
        ) == MatrixConstraint.StrictLowerBound

    @property
    def strict_upper(self):
        """A boolean indicating whether this constraint has a strict upper bound."""
        return (
            self.parent_component()._range_types[self.index()]
            & MatrixConstraint.StrictUpperBound
        ) == MatrixConstraint.StrictUpperBound

    def set_value(self, expr):
        """Set the expression on this constraint."""
        raise NotImplementedError("MatrixConstraint row elements can not be updated")


@ModelComponentFactory.register("A set of constraint expressions in Ax=b form.")
class MatrixConstraint(Mapping, IndexedConstraint):
    #
    # Bound types
    # (make sure the maximum value here
    #  will fit in an unsigned char)
    #
    StrictUpperBound = 0b00011
    UpperBound = 0b00010
    Equality = 0b01110
    LowerBound = 0b01000
    StrictLowerBound = 0b11000
    NoBound = 0b00000

    def __init__(
        self, nrows, ncols, nnz, prows, jcols, vals, ranges, range_types, varmap
    ):
        assert len(prows) == nrows + 1
        assert len(jcols) == nnz
        assert len(vals) == nnz
        assert len(ranges) == 2 * nrows
        assert len(range_types) == nrows
        assert len(varmap) == ncols

        IndexedConstraint.__init__(self, Any)

        self._prows = prows
        self._jcols = jcols
        self._vals = vals
        self._ranges = ranges
        self._range_types = range_types
        self._varmap = varmap

    def construct(self, data=None):
        """
        Construct the expression(s) for this constraint.
        """
        if is_debug_set(logger):
            logger.debug("Constructing constraint %s" % (self.name))
        if self._constructed:
            return
        self._constructed = True

        _init = _LinearMatrixConstraintData
        self._data = tuple(
            _init(i, component=self) for i in range(len(self._range_types))
        )

    #
    # Override some IndexedComponent methods
    #

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return self._data.__len__()

    def __iter__(self):
        return iter(i for i in range(len(self)))

    #
    # Remove methods that allow modifying this constraint
    #

    def add(self, index, expr):
        raise NotImplementedError

    def __delitem__(self):
        raise NotImplementedError

    #
    # Pyomo components support an extended dict API
    #
    def keys(self, sort=None):
        # The 0..n-1 indices are always ordered and sorted; we can
        # ignore the `sort` argument
        return super().keys()

    def values(self, sort=None):
        # The 0..n-1 indices are always ordered and sorted; we can
        # ignore the `sort` argument
        return super().values()

    def items(self, sort=None):
        # The 0..n-1 indices are always ordered and sorted; we can
        # ignore the `sort` argument
        return super().items()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.contrib.mpc.interfaces.var_linker import DynamicVarLinker


class TestVarLinker(unittest.TestCase):
    def _make_models(self, n_time_points_1=3, n_time_points_2=3):
        m1 = pyo.ConcreteModel()
        m1.time = pyo.Set(initialize=range(n_time_points_1))
        m1.comp = pyo.Set(initialize=["A", "B"])
        m1.var = pyo.Var(
            m1.time,
            m1.comp,
            initialize={(i, j): 1.0 + i * 0.1 for i, j in m1.time * m1.comp},
        )
        m1.input = pyo.Var(m1.time, initialize={i: 1.0 - i * 0.1 for i in m1.time})

        m2 = pyo.ConcreteModel()
        m2.time = pyo.Set(initialize=range(n_time_points_2))
        m2.x1 = pyo.Var(m2.time, initialize=2.1)
        m2.x2 = pyo.Var(m2.time, initialize=2.2)
        m2.x3 = pyo.Var(m2.time, initialize=2.3)
        m2.x4 = pyo.Var(m2.time, initialize=2.4)

        return m1, m2

    def test_transfer_one_to_one(self):
        m1, m2 = self._make_models()
        vars1 = [pyo.Reference(m1.var[:, "A"]), pyo.Reference(m1.var[:, "B"]), m1.input]
        vars2 = [m2.x1, m2.x2, m2.x3]

        linker = DynamicVarLinker(vars1, vars2)
        t_source = 0
        t_target = 2
        linker.transfer(t_source=0, t_target=2)

        pred_AB = lambda t: 1.0 + t * 0.1
        pred_input = lambda t: 1.0 - t * 0.1
        for t in m1.time:
            # Both models have same time set

            # Values in source variables have not changed
            self.assertEqual(m1.var[t, "A"].value, pred_AB(t))
            self.assertEqual(m1.var[t, "B"].value, pred_AB(t))
            self.assertEqual(m1.input[t].value, pred_input(t))

            if t == t_target:
                self.assertEqual(m2.x1[t].value, pred_AB(t_source))
                self.assertEqual(m2.x2[t].value, pred_AB(t_source))
                self.assertEqual(m2.x3[t].value, pred_input(t_source))
                self.assertEqual(m2.x4[t].value, 2.4)
            else:
                self.assertEqual(m2.x1[t].value, 2.1)
                self.assertEqual(m2.x2[t].value, 2.2)
                self.assertEqual(m2.x3[t].value, 2.3)
                self.assertEqual(m2.x4[t].value, 2.4)

    def test_transfer_one_to_all(self):
        m1, m2 = self._make_models()
        vars1 = [pyo.Reference(m1.var[:, "A"]), pyo.Reference(m1.var[:, "B"]), m1.input]
        vars2 = [m2.x1, m2.x2, m2.x3]

        linker = DynamicVarLinker(vars1, vars2)
        t_source = 0
        t_target = 2
        linker.transfer(t_source=0, t_target=m2.time)

        pred_AB = lambda t: 1.0 + t * 0.1
        pred_input = lambda t: 1.0 - t * 0.1
        for t in m1.time:
            # Both models have same time set

            # Values in source variables have not changed
            self.assertEqual(m1.var[t, "A"].value, pred_AB(t))
            self.assertEqual(m1.var[t, "B"].value, pred_AB(t))
            self.assertEqual(m1.input[t].value, pred_input(t))

            # Target variables have been updated
            self.assertEqual(m2.x1[t].value, pred_AB(t_source))
            self.assertEqual(m2.x2[t].value, pred_AB(t_source))
            self.assertEqual(m2.x3[t].value, pred_input(t_source))
            self.assertEqual(m2.x4[t].value, 2.4)

    def test_transfer_all_to_all(self):
        m1, m2 = self._make_models()
        vars1 = [pyo.Reference(m1.var[:, "A"]), pyo.Reference(m1.var[:, "B"]), m1.input]
        vars2 = [m2.x1, m2.x2, m2.x3]

        linker = DynamicVarLinker(vars1, vars2)
        t_source = 0
        t_target = 2
        linker.transfer(t_source=m1.time, t_target=m2.time)

        pred_AB = lambda t: 1.0 + t * 0.1
        pred_input = lambda t: 1.0 - t * 0.1
        for t in m1.time:
            # Both models have same time set

            # Values in source variables have not changed
            self.assertEqual(m1.var[t, "A"].value, pred_AB(t))
            self.assertEqual(m1.var[t, "B"].value, pred_AB(t))
            self.assertEqual(m1.input[t].value, pred_input(t))

            # Target variables have been updated
            self.assertEqual(m2.x1[t].value, pred_AB(t))
            self.assertEqual(m2.x2[t].value, pred_AB(t))
            self.assertEqual(m2.x3[t].value, pred_input(t))
            self.assertEqual(m2.x4[t].value, 2.4)

    def test_transfer_exceptions(self):
        m1, m2 = self._make_models()
        vars1 = [pyo.Reference(m1.var[:, "A"]), pyo.Reference(m1.var[:, "B"])]
        vars2 = [m2.x1, m2.x2, m2.x3]

        msg = "must be provided two lists.*of equal length"
        with self.assertRaisesRegex(ValueError, msg):
            linker = DynamicVarLinker(vars1, vars2)

        vars1 = [pyo.Reference(m1.var[:, "A"]), pyo.Reference(m1.var[:, "B"]), m1.input]
        vars2 = [m2.x1, m2.x2, m2.x3]
        linker = DynamicVarLinker(vars1, vars2)
        msg = "Source time points were not provided"
        with self.assertRaisesRegex(RuntimeError, msg):
            linker.transfer(t_target=m2.time)

        msg = "Target time points were not provided"
        with self.assertRaisesRegex(RuntimeError, msg):
            linker.transfer(t_source=m1.time.first())


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for ComponentUID
#
import pickle
from collections import namedtuple
from datetime import datetime
from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Block,
    Var,
    Set,
    Param,
    Constraint,
    Any,
    ComponentUID,
    Reference,
)
from pyomo.core.base.indexed_component import IndexedComponent
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from pyomo.common.log import LoggingIntercept

_star = slice(None)

_Foo = namedtuple('_Foo', ['x', 'yy'])


class TestComponentUID(unittest.TestCase):
    def setUp(self):
        self.m = ConcreteModel()
        m = self.m
        m.a = Param()
        m.s = Set(initialize=[1, '2', 3])
        m.b = Block(m.s, m.s)
        m.b[1, 1].c = Block()
        m.b[1, '2'].c = Block()
        m.b[1, '2'].c.a = Param(m.s, initialize=3, mutable=True)

    def tearDown(self):
        self.m = None

    def test_genFromComponent_simple(self):
        cuid = ComponentUID(self.m.a)
        self.assertEqual(cuid._cids, (('a', tuple()),))

    def test_genFromComponent_nested(self):
        cuid = ComponentUID(self.m.b[1, '2'].c.a[3])
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (3,))))

    def test_genFromComponent_indexed(self):
        cuid = ComponentUID(self.m.b[1, '2'].c.a)
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', ())))

    def test_genFromComponent_nameBuffer(self):
        buf = {}
        cuid = ComponentUID(self.m.b[1, '2'].c.a, cuid_buffer=buf)
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', ())))
        self.assertEqual(len(buf), 9)
        for s1 in self.m.s:
            for s2 in self.m.s:
                _id = id(self.m.b[s1, s2])
                self.assertIn(_id, buf)
                self.assertEqual(buf[_id], ('b', (s1, s2)))

    def test_genFromComponent_context(self):
        cuid = ComponentUID(self.m.b[1, '2'].c.a, context=self.m.b[1, '2'])
        self.assertEqual(cuid._cids, (('c', tuple()), ('a', ())))
        with self.assertRaisesRegex(
            ValueError, r"Context 'b\[1,'2'\]' does not apply to component 's'"
        ):
            ComponentUID(self.m.s, context=self.m.b[1, '2'])
        with self.assertRaisesRegex(ValueError, "Context is not allowed"):
            ComponentUID("b[1,2].c.a[2]", context=self.m.b[1, '2'])

    def test_parseFromString(self):
        cuid = ComponentUID('b[1,2].c.a[2]')
        self.assertEqual(cuid._cids, (('b', (1, 2)), ('c', tuple()), ('a', (2,))))

    def test_parseFromString_singleQuote(self):
        cuid = ComponentUID('b[1,\'2\'].c.a[2]')
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (2,))))

    def test_parseFromString_doubleQuote(self):
        cuid = ComponentUID('b[1,\"2\"].c.a[2]')
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (2,))))

    def test_parseFromString_typeID(self):
        cuid = ComponentUID('b[#1,$2].c.a[2]')
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (2,))))

    def test_parseFromString_wildcard_1(self):
        cuid = ComponentUID('b[**].c.a[*]')
        self.assertEqual(
            cuid._cids, (('b', (Ellipsis,)), ('c', tuple()), ('a', (_star,)))
        )

    def test_parseFromString_wildcard_2(self):
        cuid = ComponentUID('b[*,*].c.a[*]')
        self.assertEqual(
            cuid._cids, (('b', (_star, _star)), ('c', tuple()), ('a', (_star,)))
        )

    def test_parseFromString_spaces(self):
        cuid = ComponentUID('x[a b,c d]')
        self.assertEqual(cuid._cids, (('x', ('a b', 'c d')),))

        cuid = ComponentUID("x['a b',\"c d\"]")
        self.assertEqual(cuid._cids, (('x', ('a b', 'c d')),))

        cuid = ComponentUID('x[a b, c d]')
        self.assertEqual(cuid._cids, (('x', ('a b', 'c d')),))

        cuid = ComponentUID("x[ a b , 'c d' ]")
        self.assertEqual(cuid._cids, (('x', ('a b', 'c d')),))

    def test_parseFromRepr1(self):
        cuid = ComponentUID('b:1,2.c.a:2')
        self.assertEqual(cuid._cids, (('b', (1, 2)), ('c', tuple()), ('a', (2,))))

    def test_parseFromRepr1_singleQuote(self):
        cuid = ComponentUID('b:1,\'2\'.c.a:2')
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (2,))))

    def test_parseFromRepr1_doubleQuote(self):
        cuid = ComponentUID('b:1,\"2\".c.a:2')
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (2,))))

    def test_parseFromRepr1_typeID(self):
        cuid = ComponentUID('b:#1,$2.c.a:2')
        self.assertEqual(cuid._cids, (('b', (1, '2')), ('c', tuple()), ('a', (2,))))

    def test_parseFromRepr1_wildcard_1(self):
        cuid = ComponentUID('b:**.c.a:*')
        self.assertEqual(
            cuid._cids, (('b', (Ellipsis,)), ('c', tuple()), ('a', (_star,)))
        )

    def test_parseFromRepr1_wildcard_2(self):
        cuid = ComponentUID('b:*,*.c.a:*')
        self.assertEqual(
            cuid._cids, (('b', (_star, _star)), ('c', tuple()), ('a', (_star,)))
        )

    def test_parseFromRepr2_lexError(self):
        cuid = ComponentUID('')  # Bogus instance to access parser
        with self.assertRaisesRegex(IOError, "ERROR: Token ':' Line 1 Column 1"):
            list(cuid._parse_cuid_v2(':'))
        with self.assertRaisesRegex(IOError, "ERROR: Token '\n].b:' Line 1 Column 3"):
            list(cuid._parse_cuid_v2('a[\n].b:'))

    def test_escapeChars(self):
        ref = r"b['a\n.b\\'].x"
        cuid = ComponentUID(ref)
        self.assertEqual(cuid._cids, (('b', ('a\n.b\\',)), ('x', tuple())))

        m = ConcreteModel()
        m.b = Block(['a\n.b\\'])
        m.b['a\n.b\\'].x = x = Var()
        self.assertTrue(cuid.matches(x))
        self.assertEqual(repr(ComponentUID(x)), ref)
        self.assertEqual(str(ComponentUID(x)), r"b['a\n.b\\'].x")

    def test_nonIntNumber(self):
        inf = float('inf')
        m = ConcreteModel()
        m.b = Block([inf, 'inf'])

        m.b[inf].x = x = Var()
        ref = r"b[inf].x"

        cuid = ComponentUID(x)
        self.assertEqual(cuid._cids, (('b', (inf,)), ('x', tuple())))

        self.assertTrue(cuid.matches(x))
        self.assertEqual(repr(ComponentUID(x)), ref)
        self.assertEqual(str(ComponentUID(x)), ref)

        cuid = ComponentUID(ref)
        self.assertEqual(cuid._cids, (('b', (inf,)), ('x', tuple())))

        self.assertTrue(cuid.matches(x))
        self.assertEqual(repr(ComponentUID(x)), ref)
        self.assertEqual(str(ComponentUID(x)), ref)

        ref = r"b:#inf.x"
        cuid = ComponentUID(ref)
        self.assertEqual(cuid._cids, (('b', (inf,)), ('x', tuple())))

        self.assertTrue(cuid.matches(x))
        self.assertEqual(ComponentUID(x).get_repr(1), ref)
        self.assertEqual(str(ComponentUID(x)), r"b[inf].x")

        #
        m.b['inf'].x = x = Var()
        ref = r"b['inf'].x"
        #

        cuid = ComponentUID(x)
        self.assertEqual(cuid._cids, (('b', ('inf',)), ('x', tuple())))

        self.assertTrue(cuid.matches(x))
        self.assertEqual(repr(ComponentUID(x)), ref)
        self.assertEqual(str(ComponentUID(x)), ref)

        cuid = ComponentUID(ref)
        self.assertEqual(cuid._cids, (('b', ('inf',)), ('x', tuple())))

        self.assertTrue(cuid.matches(x))
        self.assertEqual(repr(ComponentUID(x)), ref)
        self.assertEqual(str(ComponentUID(x)), ref)

        ref = r"b:$inf.x"
        cuid = ComponentUID(ref)
        self.assertEqual(cuid._cids, (('b', ('inf',)), ('x', tuple())))

        self.assertTrue(cuid.matches(x))
        self.assertEqual(ComponentUID(x).get_repr(1), ref)
        self.assertEqual(str(ComponentUID(x)), r"b['inf'].x")

    def test_find_component_deprecated(self):
        ref = self.m.b[1, '2'].c.a[3]
        cuid = ComponentUID(ref)
        DEP_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo.core'):
            self.assertTrue(cuid.find_component(self.m) is ref)
        self.assertIn(
            'ComponentUID.find_component() is deprecated.', DEP_OUT.getvalue()
        )

    def test_find_explicit_exists(self):
        ref = self.m.b[1, '2'].c.a[3]
        cuid = ComponentUID(ref)
        self.assertTrue(cuid.find_component_on(self.m) is ref)

    def test_find_component_exists_1(self):
        ref = self.m.b[1, '2'].c.a
        cuid = ComponentUID(ref)
        self.assertTrue(cuid.find_component_on(self.m) is ref)

    def test_find_wildcard(self):
        cuid = ComponentUID('b:1,$2.c.a:*')
        comp = cuid.find_component_on(self.m)
        self.assertIs(comp.ctype, Param)
        cList = list(comp.values())
        self.assertEqual(len(cList), 3)
        self.assertEqual(cList, list(self.m.b[1, '2'].c.a[:]))

        cuid = ComponentUID('b[*,*]')
        comp = cuid.find_component_on(self.m)
        self.assertIs(comp.ctype, Block)
        cList = list(comp.values())
        self.assertEqual(len(cList), 9)
        self.assertEqual(cList, list(self.m.b.values()))

    def test_find_wildcard_partial_exists(self):
        # proper Reference: to ComponentData
        cuid = ComponentUID('b[*,*].c.a[**]')
        comp = cuid.find_component_on(self.m)
        self.assertIs(comp.ctype, Param)
        cList = list(comp.values())
        self.assertEqual(len(cList), 3)
        self.assertEqual(cList, list(self.m.b[1, '2'].c.a[:]))

        # improper Reference: to IndexedComponent
        cuid = ComponentUID('b[*,*].c.a')
        comp = cuid.find_component_on(self.m)
        self.assertIs(comp.ctype, IndexedComponent)
        cList = list(comp.values())
        self.assertEqual(len(cList), 1)
        self.assertIs(cList[0], self.m.b[1, '2'].c.a)

    def test_find_wildcard_not_exists(self):
        cuid = ComponentUID('b[*,*].c.x')
        self.assertIsNone(cuid.find_component_on(self.m))

    # def test_find_implicit_exists(self):
    #     cuid = ComponentUID('b:1,2.c.a:3')
    #     self.assertTrue(cuid.find_component_on(self.m) is
    #                     self.m.b[1,'2'].c.a[3])

    def test_find_implicit_notExists_1(self):
        cuid = ComponentUID('b:1,2.c.a:4')
        self.assertTrue(cuid.find_component_on(self.m) is None)

    def test_find_implicit_notExists_2(self):
        cuid = ComponentUID('b:1,1.c.a:3')
        self.assertTrue(cuid.find_component_on(self.m) is None)

    def test_find_explicit_notExists_1(self):
        cuid = ComponentUID('b:1,2.c.a:$3')
        self.assertTrue(cuid.find_component_on(self.m) is None)

    def test_find_explicit_notExists_2(self):
        cuid = ComponentUID('b:$1,2.c.a:3')
        self.assertTrue(cuid.find_component_on(self.m) is None)

    def test_printers_1(self):
        cuid = ComponentUID(self.m.b[1, '2'].c.a[3])
        s = "b[1,'2'].c.a[3]"
        r1 = "b:#1,$2.c.a:#3"
        r2 = "b[1,'2'].c.a[3]"
        self.assertEqual(str(cuid), s)
        self.assertEqual(repr(cuid), r2)
        self.assertEqual(cuid.get_repr(1), r1)
        self.assertEqual(cuid.get_repr(2), r2)
        with self.assertRaisesRegex(
            ValueError, "Invalid repr version '3'; expected 1 or 2"
        ):
            cuid.get_repr(3)

    def test_printers_2(self):
        cuid = ComponentUID('b:$1,2.c.a:#3')
        s = "b['1',2].c.a[3]"
        r1 = "b:$1,#2.c.a:#3"
        r2 = "b['1',2].c.a[3]"
        self.assertEqual(str(cuid), s)
        self.assertEqual(repr(cuid), r2)
        self.assertEqual(cuid.get_repr(1), r1)
        self.assertEqual(cuid.get_repr(2), r2)

    def test_printers_3(self):
        cuid = ComponentUID('b:**.c.a:*')
        s = 'b[**].c.a[*]'
        r1 = "b:**.c.a:*"
        r2 = "b[**].c.a[*]"
        self.assertEqual(str(cuid), s)
        self.assertEqual(repr(cuid), r2)
        self.assertEqual(cuid.get_repr(1), r1)
        self.assertEqual(cuid.get_repr(2), r2)

    def test_printers_4(self):
        cuid = ComponentUID('b:*,*.c.a:**')
        s = 'b[*,*].c.a[**]'
        r1 = "b:*,*.c.a:**"
        r2 = "b[*,*].c.a[**]"
        self.assertEqual(str(cuid), s)
        self.assertEqual(repr(cuid), r2)
        self.assertEqual(cuid.get_repr(1), r1)
        self.assertEqual(cuid.get_repr(2), r2)

    def test_matches_explicit(self):
        cuid = ComponentUID(self.m.b[1, '2'].c.a[3])
        self.assertTrue(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a['2']))

    # def test_matches_implicit(self):
    #     cuid = ComponentUID('b:1,2.c.a:3')
    #     self.assertTrue(cuid.matches(self.m.b[1,'2'].c.a[3]))
    #     self.assertFalse(cuid.matches(self.m.b[1,'2'].c.a['2']))

    def test_matches_explicit_1(self):
        cuid = ComponentUID('b:#1,$2.c.a:$3')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_explicit_2(self):
        cuid = ComponentUID('b:#1,#2.c.a:#3')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_wildcard_1(self):
        cuid = ComponentUID('b:**.c.a:*')
        self.assertTrue(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertTrue(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_wildcard_2(self):
        cuid = ComponentUID('b:*,*.c.a:**')
        self.assertTrue(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertTrue(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_wildcard_3(self):
        cuid = ComponentUID('b:*,*.c.a:*,*')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_mismatch_name(self):
        cuid = ComponentUID('b:*,*.d')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))

    def test_matches_mismatch_1(self):
        cuid = ComponentUID('b:*,*.c.a:*')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))

    def test_matches_mismatch_2(self):
        cuid = ComponentUID('b:*,*.c')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_mismatch_3(self):
        cuid = ComponentUID('b:*,*,*.c.a:*')
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a[3]))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c.a['2']))

    def test_matches_ellipsis1(self):
        cuid = ComponentUID('b[**,1].c')
        self.assertTrue(cuid.matches(self.m.b[1, 1].c))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))

    def test_matches_ellipsis2(self):
        cuid = ComponentUID('b[**,1,1].c')
        self.assertTrue(cuid.matches(self.m.b[1, 1].c))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))

    def test_matches_ellipsis3(self):
        cuid = ComponentUID('b[**,1,1,3].c')
        self.assertFalse(cuid.matches(self.m.b[1, 1].c))
        self.assertFalse(cuid.matches(self.m.b[1, '2'].c))

    def test_matches_ellipsis4(self):
        cuid = ComponentUID('b[**,1,*].c')
        self.assertTrue(cuid.matches(self.m.b[1, 1].c))
        self.assertTrue(cuid.matches(self.m.b[1, '2'].c))

    def test_list_components_done_1(self):
        cuid = ComponentUID('b:*,*,*.c.a:*')
        ref = []
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_list_components_done_2(self):
        cuid = ComponentUID('b:*,*.c:#1.a:*')
        ref = []
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_list_components_scalar(self):
        cuid = ComponentUID('b:1,$2.c.a:3')
        ref = [str(ComponentUID(self.m.b[1, '2'].c.a[3]))]
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_list_components_wildcard_1(self):
        cuid = ComponentUID('b:**.c.a:3')
        ref = [str(ComponentUID(self.m.b[1, '2'].c.a[3]))]
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_list_components_wildcard_2(self):
        cuid = ComponentUID('b:*,*.c.a:*')
        ref = [
            str(ComponentUID(self.m.b[1, '2'].c.a[1])),
            str(ComponentUID(self.m.b[1, '2'].c.a['2'])),
            str(ComponentUID(self.m.b[1, '2'].c.a[3])),
        ]
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_list_components_wildcard_3(self):
        cuid = ComponentUID('b:1,*.c')
        ref = [
            str(ComponentUID(self.m.b[1, 1].c)),
            str(ComponentUID(self.m.b[1, '2'].c)),
        ]
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_list_components_wildcard_4(self):
        cuid = ComponentUID('b:1,*')
        ref = [
            str(ComponentUID(self.m.b[1, 1])),
            str(ComponentUID(self.m.b[1, '2'])),
            str(ComponentUID(self.m.b[1, 3])),
        ]
        cList = [str(ComponentUID(x)) for x in cuid.list_components(self.m)]
        self.assertEqual(sorted(cList), sorted(ref))

    def test_in_container(self):
        a = ComponentUID('foo.bar[*]')
        b = ComponentUID('baz')
        c = ComponentUID('baz.bar')
        D = {a: 1, b: 2}

        self.assertTrue(a in D)
        self.assertTrue(b in D)
        self.assertFalse(c in D)

        # Verify that hashing is not being done by id()
        self.assertTrue(ComponentUID('foo.bar[*]') in D)
        self.assertTrue(ComponentUID('baz') in D)

    def test_comparisons(self):
        a = ComponentUID('foo.x[*]')
        b = ComponentUID('baz')

        self.assertFalse(a < b)
        self.assertFalse(a <= b)
        self.assertTrue(a > b)
        self.assertTrue(a >= b)
        self.assertFalse(a == b)
        self.assertTrue(a != b)

        self.assertTrue(b < a)
        self.assertTrue(b <= a)
        self.assertFalse(b > a)
        self.assertFalse(b >= a)
        self.assertFalse(b == a)
        self.assertTrue(b != a)

        self.assertFalse(a < ComponentUID('baz'))
        self.assertFalse(a <= ComponentUID('baz'))
        self.assertTrue(a > ComponentUID('baz'))
        self.assertTrue(a >= ComponentUID('baz'))
        self.assertFalse(a == ComponentUID('baz'))
        self.assertTrue(a != ComponentUID('baz'))

        self.assertTrue(ComponentUID('baz') < a)
        self.assertTrue(ComponentUID('baz') <= a)
        self.assertFalse(ComponentUID('baz') > a)
        self.assertFalse(ComponentUID('baz') >= a)
        self.assertFalse(ComponentUID('baz') == a)
        self.assertTrue(ComponentUID('baz') != a)

        self.assertFalse(b < b)
        self.assertTrue(b <= b)
        self.assertFalse(b > b)
        self.assertTrue(b >= b)
        self.assertTrue(b == b)
        self.assertFalse(b != b)

        self.assertFalse(ComponentUID('baz') < b)
        self.assertTrue(ComponentUID('baz') <= b)
        self.assertFalse(ComponentUID('baz') > b)
        self.assertTrue(ComponentUID('baz') >= b)
        self.assertTrue(ComponentUID('baz') == b)
        self.assertFalse(ComponentUID('baz') != b)

    def test_comparisons_lt(self):
        a = ComponentUID('foo.x[*]')
        a1 = ComponentUID('foo.x[1]')
        a2 = ComponentUID('foo.x[2]')
        aa = ComponentUID("foo.x['a']")
        a11 = ComponentUID('foo.x[1,1]')
        ae = ComponentUID('foo.x[**]')
        self.assertTrue(a < ae)
        self.assertTrue(a1 < ae)
        self.assertTrue(a1 < a)
        self.assertTrue(a1 < a2)
        self.assertTrue(a1 < aa)
        self.assertTrue(a1 < a11)
        self.assertTrue(a11 < a2)
        self.assertFalse(ae < a)
        self.assertFalse(ae < a1)
        self.assertFalse(a < a1)
        self.assertFalse(a2 < a1)
        self.assertFalse(aa < a1)
        self.assertFalse(a11 < a1)
        self.assertFalse(a2 < a11)

        x = ComponentUID('foo.x')
        xy = ComponentUID('foo.x.y')
        self.assertTrue(x < xy)
        self.assertFalse(xy < x)

        with self.assertRaisesRegex(
            TypeError, "'<' not supported between instances of 'ComponentUID' and 'int'"
        ):
            a < 5

    def test_comparisons_eq(self):
        a = ComponentUID('foo.x[*]')
        a1 = ComponentUID('foo.x[1]')
        b = ComponentUID('foo.x[*]')
        self.assertEqual(a, b)
        self.assertNotEqual(a, a1)
        self.assertNotEqual(a, 5)

    def test_generate_cuid_string_map(self):
        model = Block(concrete=True)
        model.x = Var()
        model.y = Var([1, 2])
        model.V = Var([('a', 'b'), (1, '2'), (3, 4)])
        model.b = Block(concrete=True)
        model.b.z = Var([1, '2'])
        setattr(model.b, '.H', Var(['a', 2]))
        model.B = Block(['a', 2], concrete=True)
        setattr(model.B['a'], '.k', Var())
        model.B[2].b = Block()
        model.B[2].b.x = Var()
        model.add_component('c tuple', Constraint(Any))
        model.component('c tuple')[(1,)] = model.x >= 0

        cuids = (
            ComponentUID.generate_cuid_string_map(model, repr_version=1),
            ComponentUID.generate_cuid_string_map(model),
        )
        self.assertEqual(len(cuids[0]), 24)
        self.assertEqual(len(cuids[1]), 24)
        for obj in [
            model,
            model.x,
            model.y,
            model.y[1],
            model.y[2],
            model.V,
            model.V['a', 'b'],
            model.V[1, '2'],
            model.V[3, 4],
            model.b,
            model.b.z,
            model.b.z[1],
            model.b.z['2'],
            getattr(model.b, '.H'),
            getattr(model.b, '.H')['a'],
            getattr(model.b, '.H')[2],
            model.B,
            model.B['a'],
            getattr(model.B['a'], '.k'),
            model.B[2],
            model.B[2].b,
            model.B[2].b.x,
            model.component('c tuple')[(1,)],
        ]:
            self.assertEqual(ComponentUID(obj).get_repr(1), cuids[0][obj])
            self.assertEqual(repr(ComponentUID(obj)), cuids[1][obj])

        cuids = (
            ComponentUID.generate_cuid_string_map(
                model, descend_into=False, repr_version=1
            ),
            ComponentUID.generate_cuid_string_map(model, descend_into=False),
        )
        self.assertEqual(len(cuids[0]), 15)
        self.assertEqual(len(cuids[1]), 15)
        for obj in [
            model,
            model.x,
            model.y,
            model.y[1],
            model.y[2],
            model.V,
            model.V['a', 'b'],
            model.V[1, '2'],
            model.V[3, 4],
            model.b,
            model.B,
            model.B['a'],
            model.B[2],
            model.component('c tuple')[(1,)],
        ]:
            self.assertEqual(ComponentUID(obj).get_repr(1), cuids[0][obj])
            self.assertEqual(repr(ComponentUID(obj)), cuids[1][obj])

        cuids = (
            ComponentUID.generate_cuid_string_map(model, ctype=Var, repr_version=1),
            ComponentUID.generate_cuid_string_map(model, ctype=Var),
        )
        self.assertEqual(len(cuids[0]), 22)
        self.assertEqual(len(cuids[1]), 22)
        for obj in [
            model,
            model.x,
            model.y,
            model.y[1],
            model.y[2],
            model.V,
            model.V['a', 'b'],
            model.V[1, '2'],
            model.V[3, 4],
            model.b,
            model.b.z,
            model.b.z[1],
            model.b.z['2'],
            getattr(model.b, '.H'),
            getattr(model.b, '.H')['a'],
            getattr(model.b, '.H')[2],
            model.B,
            model.B['a'],
            getattr(model.B['a'], '.k'),
            model.B[2],
            model.B[2].b,
            model.B[2].b.x,
        ]:
            self.assertEqual(ComponentUID(obj).get_repr(1), cuids[0][obj])
            self.assertEqual(repr(ComponentUID(obj)), cuids[1][obj])

        cuids = (
            ComponentUID.generate_cuid_string_map(
                model, ctype=Var, descend_into=False, repr_version=1
            ),
            ComponentUID.generate_cuid_string_map(model, ctype=Var, descend_into=False),
        )
        self.assertEqual(len(cuids[0]), 9)
        self.assertEqual(len(cuids[1]), 9)
        for obj in [
            model,
            model.x,
            model.y,
            model.y[1],
            model.y[2],
            model.V,
            model.V['a', 'b'],
            model.V[1, '2'],
            model.V[3, 4],
        ]:
            self.assertEqual(ComponentUID(obj).get_repr(1), cuids[0][obj])
            self.assertEqual(repr(ComponentUID(obj)), cuids[1][obj])

    def test_pickle(self):
        a = ComponentUID("b[1,'2'].c")
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a, b)

        a = ComponentUID("b[1,*].c")
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a, b)

        a = ComponentUID("b[**,*].c")
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a, b)

    def test_findComponentOn_nestedTuples(self):
        # Tests for #1069
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(Any)
        m.c[0] = m.x >= 0
        m.c[(1,)] = m.x >= 1
        m.c[(2,)] = m.x >= 2
        m.c[2] = m.x >= 3
        self.assertIs(ComponentUID(m.c[0]).find_component_on(m), m.c[0])
        self.assertIs(ComponentUID('c[0]').find_component_on(m), m.c[0])
        self.assertIsNone(ComponentUID('c[(0,)]').find_component_on(m))
        self.assertIs(ComponentUID(m.c[(1,)]).find_component_on(m), m.c[(1,)])
        self.assertIs(ComponentUID('c[(1,)]').find_component_on(m), m.c[(1,)])
        self.assertIsNone(ComponentUID('c[1]').find_component_on(m))
        self.assertIs(ComponentUID('c[(2,)]').find_component_on(m), m.c[(2,)])
        self.assertIs(ComponentUID('c[2]').find_component_on(m), m.c[2])
        self.assertEqual(len(m.c), 4)

        self.assertEqual(repr(ComponentUID(m.c[0])), "c[0]")
        self.assertEqual(repr(ComponentUID(m.c[(1,)])), "c[(1,)]")
        self.assertEqual(str(ComponentUID(m.c[0])), "c[0]")
        self.assertEqual(str(ComponentUID(m.c[(1,)])), "c[(1,)]")

        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint([0, 1])
        m.c[0] = m.x >= 0
        m.c[(1,)] = m.x >= 1
        self.assertIs(ComponentUID(m.c[0]).find_component_on(m), m.c[0])
        self.assertIs(ComponentUID(m.c[(0,)]).find_component_on(m), m.c[0])
        self.assertIs(ComponentUID('c[0]').find_component_on(m), m.c[0])
        self.assertIs(ComponentUID('c[(0,)]').find_component_on(m), m.c[0])
        self.assertIs(ComponentUID(m.c[1]).find_component_on(m), m.c[1])
        self.assertIs(ComponentUID(m.c[(1,)]).find_component_on(m), m.c[1])
        self.assertIs(ComponentUID('c[(1,)]').find_component_on(m), m.c[1])
        self.assertIs(ComponentUID('c[1]').find_component_on(m), m.c[1])
        self.assertEqual(len(m.c), 2)

        m = ConcreteModel()
        m.b = Block(Any)
        m.b[0].c = Block(Any)
        m.b[0].c[0].x = Var()
        m.b[(1,)].c = Block(Any)
        m.b[(1,)].c[(1,)].x = Var()
        ref = m.b[0].c[0].x
        self.assertIs(ComponentUID(ref).find_component_on(m), ref)
        ref = 'm.b[0].c[(0,)].x'
        self.assertIsNone(ComponentUID(ref).find_component_on(m))
        ref = m.b[(1,)].c[(1,)].x
        self.assertIs(ComponentUID(ref).find_component_on(m), ref)
        ref = 'm.b[(1,)].c[1].x'
        self.assertIsNone(ComponentUID(ref).find_component_on(m))

        buf = {}
        ref = m.b[0].c[0].x
        self.assertIs(ComponentUID(ref, cuid_buffer=buf).find_component_on(m), ref)
        self.assertEqual(len(buf), 3)
        ref = 'm.b[0].c[(0,)].x'
        self.assertIsNone(ComponentUID(ref, cuid_buffer=buf).find_component_on(m))
        self.assertEqual(len(buf), 3)
        ref = m.b[(1,)].c[(1,)].x
        self.assertIs(ComponentUID(ref, cuid_buffer=buf).find_component_on(m), ref)
        self.assertEqual(len(buf), 4)
        ref = 'm.b[(1,)].c[1].x'
        self.assertIsNone(ComponentUID(ref, cuid_buffer=buf).find_component_on(m))
        self.assertEqual(len(buf), 4)

    def test_pickle_index(self):
        m = ConcreteModel()
        m.b = Block(Any)

        idx = "|b'foo'"
        m.b[idx].x = Var()
        cuid = ComponentUID(m.b[idx].x)
        self.assertEqual(str(cuid), 'b["|b\'foo\'"].x')
        self.assertIs(cuid.find_component_on(m), m.b[idx].x)
        tmp = ComponentUID(str(cuid))
        self.assertIsNot(cuid, tmp)
        self.assertEqual(cuid, tmp)
        self.assertIs(tmp.find_component_on(m), m.b[idx].x)
        tmp = pickle.loads(pickle.dumps(cuid))
        self.assertIsNot(cuid, tmp)
        self.assertEqual(cuid, tmp)
        self.assertIs(tmp.find_component_on(m), m.b[idx].x)

        idx = _Foo(1, 'a')
        m.b[idx].x = Var()
        cuid = ComponentUID(m.b[idx].x)
        # Note that the pickle string for namedtuple changes between
        # Python 2, 3, and pypy, so we will just check the non-pickle
        # data part
        self.assertRegex(str(cuid), r"^b\[\|b?(['\"]).*\.\1\]\.x$")

        self.assertIs(cuid.find_component_on(m), m.b[idx].x)
        tmp = ComponentUID(str(cuid))
        self.assertIsNot(cuid, tmp)
        self.assertEqual(cuid, tmp)
        self.assertIs(tmp.find_component_on(m), m.b[idx].x)
        tmp = pickle.loads(pickle.dumps(cuid))
        self.assertIsNot(cuid, tmp)
        self.assertEqual(cuid, tmp)
        self.assertIs(tmp.find_component_on(m), m.b[idx].x)

        idx = datetime(1, 2, 3)
        m.b[idx].x = Var()
        cuid = ComponentUID(m.b[idx].x)
        # Note that the pickle string for namedtuple changes between
        # Python 2, 3, and pypy, so we will just check the non-pickle
        # data part
        self.assertRegex(str(cuid), r"^b\[\|b?(['\"]).*\.\1\]\.x$")
        self.assertIs(cuid.find_component_on(m), m.b[idx].x)
        tmp = ComponentUID(str(cuid))
        self.assertIsNot(cuid, tmp)
        self.assertEqual(cuid, tmp)
        self.assertIs(tmp.find_component_on(m), m.b[idx].x)
        tmp = pickle.loads(pickle.dumps(cuid))
        self.assertIsNot(cuid, tmp)
        self.assertEqual(cuid, tmp)
        self.assertIs(tmp.find_component_on(m), m.b[idx].x)

        self.assertEqual(len(m.b), 3)

    def test_deprecated_ComponentUID_location(self):
        import pyomo.core.base.component as comp

        self.assertNotIn('ComponentUID', dir(comp))

        warning = (
            "DEPRECATED: the 'ComponentUID' class has been moved to "
            "'pyomo.core.base.componentuid.ComponentUID'"
        )
        OUT = StringIO()
        with LoggingIntercept(OUT, 'pyomo.core'):
            from pyomo.core.base.component import ComponentUID as old_ComponentUID
        self.assertIn(warning, OUT.getvalue().replace('\n', ' '))

        self.assertIs(old_ComponentUID, ComponentUID)
        self.assertIs(old_ComponentUID, ComponentUID)

        OUT = StringIO()
        with LoggingIntercept(OUT, 'pyomo.core'):
            self.assertIs(comp.ComponentUID, ComponentUID)
        self.assertEqual("", OUT.getvalue())

    def _slice_model(self):
        m = ConcreteModel()

        m.d1_1 = Set(initialize=[1, 2, 3])
        m.d1_2 = Set(initialize=['a', 'b', 'c'])
        m.d1_3 = Set(initialize=[1.1, 1.2, 1.3])
        m.d2 = Set(initialize=[('a', 1), ('b', 2)])
        m.dn = Set(initialize=[('c', 3), ('d', 4, 5)], dimen=None)

        @m.Block()
        def b(b):
            b.b = Block()

            @b.Block(m.d1_1)
            def b1(b1, i):
                b1.v = Var()
                b1.v1 = Var(m.d1_3)
                b1.v2 = Var(m.d1_1, m.d1_2)
                b1.vn = Var(m.dn, m.d1_2)

            @b.Block(m.d1_1, m.d1_2)
            def b2(b2, i, j):
                b2.v = Var()
                b2.v1 = Var(m.d1_3)
                b2.v2 = Var(m.d1_1, m.d1_2)
                b2.vn = Var(m.d1_1, m.dn, m.d1_2)

            @b.Block(m.d1_3, m.d2)
            def b3(b3, i, j, k):
                b3.v = Var()
                b3.v1 = Var(m.d1_3)
                b3.v2 = Var(m.d1_1, m.d1_2)
                b3.vn = Var(m.d1_1, m.dn, m.d2)

            # Don't think I can define a dim-None Block with
            # a rule unless normalize_index.flatten is False.
            b.bn = Block(m.d1_2, m.dn, m.d2)
            # NOTE: These blocks are only defined for 'a', ('a',1)
            # in the first and last "subsets"
            b.bn['a', 'c', 3, 'a', 1].v = Var()
            b.bn['a', 'c', 3, 'a', 1].v1 = Var(m.d1_3)
            b.bn['a', 'c', 3, 'a', 1].v2 = Var(m.d1_1, m.d1_2)
            b.bn['a', 'c', 3, 'a', 1].vn = Var(m.d1_1, m.dn, m.d2)
            b.bn['a', 'd', 4, 5, 'a', 1].v = Var()
            b.bn['a', 'd', 4, 5, 'a', 1].v1 = Var(m.d1_3)
            b.bn['a', 'd', 4, 5, 'a', 1].v2 = Var(m.d1_1, m.d1_2)
            b.bn['a', 'd', 4, 5, 'a', 1].vn = Var(m.d1_1, m.dn, m.d2)

        return m

    def assertListSameComponents(self, m, cuid1, cuid2):
        self.assertTrue(cuid1.list_components(m))
        self.assertEqual(
            len(list(cuid1.list_components(m))), len(list(cuid2.list_components(m)))
        )
        for c1, c2 in zip(cuid1.list_components(m), cuid2.list_components(m)):
            self.assertIs(c1, c2)

    def test_cuid_from_slice_1(self):
        """
        These are slices over a single level of the hierarchy.
        """
        m = self._slice_model()

        _slice = m.b[:]
        cuid_str = ComponentUID('b[*]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)

        _slice = m.b.b1[:]
        cuid_str = ComponentUID('b.b1[*]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)

        _slice = m.b.b1[...]
        cuid_str = ComponentUID('b.b1[**]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)

        _slice = m.b.b2[:, 'a']
        cuid_str = ComponentUID('b.b2[*,a]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[...]
        cuid_str = ComponentUID('b.b2[**]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)

        _slice = m.b.b3[1.1, :, 2]
        cuid_str = ComponentUID('b.b3[1.1,*,2]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[:, :, 'b']
        cuid_str = ComponentUID('b.b3[*,*,b]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[1.1, ...]
        cuid_str = ComponentUID('b.b3[1.1,**]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[...]
        cuid_str = ComponentUID('b.b3[**]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))

        _slice = m.b.bn['a', :, :, 'a', 1]
        cuid_str = ComponentUID('b.bn[a,*,*,a,1]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.bn['a', 'c', 3, :, :]
        cuid_str = ComponentUID('b.bn[a,c,3,*,*]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.bn[...]
        cuid_str = ComponentUID('b.bn[**]')
        cuid = ComponentUID(_slice)
        self.assertEqual(cuid, cuid_str)

    def test_cuid_from_slice_2(self):
        """
        These are slices that describe a component
        at a "deeper level" than the original slice.
        """
        m = self._slice_model()

        _slice = m.b[:].b
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b[*].b')
        self.assertEqual(cuid, cuid_str)

        _slice = m.b[:].b1[:].v
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b[*].b1[*].v')
        self.assertEqual(cuid, cuid_str)

        _slice = m.b.b2[2, :].v
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].v')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[2, :].v1[:]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].v1[*]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[2, :].v1[1.1]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].v1[1.1]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertEqual(cuid, cuid_str)
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[2, :].vn[1, ..., :, 'b']
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].vn[1,**,*,b]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[2, :].vn[..., 'b']
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].vn[**,b]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[2, :].vn[..., ...]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].vn[**,**]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[2, :].vn[...]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[2,*].vn[**]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[...].v2[:, 'a']
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[**].v2[*,a]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[:, 'a', :].v1
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b3[*,a,*].v1')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[:, 'a', :].v2[1, 'a']
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b3[*,a,*].v2[1,a]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[:, 'a', :].v2[1, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b3[*,a,*].v2[1,*]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b3[:, 'a', :].vn[1, :, :, 'a', 1]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b3[*,a,*].vn[1,*,*,a,1]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.bn['a', 'c', 3, :, :].vn[1, :, 3, 'a', :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.bn[a,c,3,*,*].vn[1,*,3,a,*]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.bn[...].vn[1, :, 3, 'a', :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.bn[**].vn[1,*,3,a,*]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.bn[...].vn
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.bn[**].vn')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.bn[...].vn[...]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.bn[**].vn[**]')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

    def test_cuid_from_slice_3(self):
        """
        "3-level" slices. These test the ability of
        the slice-processing logic to handle multiple
        `get_item` calls in a hierarchy.
        """
        m = self._slice_model()

        _slice = m.b[:].b3[:, 'a', :].v2[1, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b[*].b3[*,a,*].v2[1,*]')
        self.assertEqual(cuid, cuid_str)
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b[:].b3[:, 'a', :].v2
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b[*].b3[*,a,*].v2')
        self.assertEqual(cuid, cuid_str)
        self.assertListSameComponents(m, cuid, cuid_str)

    def test_cuid_from_slice_with_call(self):
        m = self._slice_model()

        _slice = m.b.component('b2')[:, 'a'].v2[1, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[*,a].v2[1,*]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        # This works as find_component is not in the
        # _call_stack of the slice.
        _slice = m.b.find_component('b2')[:, 'a'].v2[1, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[*,a].v2[1,*]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b[:].component('b2')
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b[*].b2')
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b[:].component('b2', 'b1')
        with self.assertRaisesRegex(ValueError, '.*multiple arguments.*'):
            cuid = ComponentUID(_slice)

        # call of something other than component
        _slice = IndexedComponent_slice(
            m.b[:].fix, (IndexedComponent_slice.call, ('fix',), {})
        )
        with self.assertRaisesRegex(
            ValueError,
            "Cannot create a CUID from a slice with a call to any "
            r"method other than 'component': got 'fix'\.",
        ):
            cuid = ComponentUID(_slice)

        _slice = IndexedComponent_slice(
            m.b[:].component('v'), (IndexedComponent_slice.call, ('fix',), {})
        )
        with self.assertRaisesRegex(
            ValueError,
            "Cannot create a CUID with a __call__ of anything "
            "other than a 'component' attribute",
        ):
            cuid = ComponentUID(_slice)

        _slice = m.b[:].component('b2', kwd=None)
        with self.assertRaisesRegex(ValueError, '.*call that contains keywords.*'):
            cuid = ComponentUID(_slice)

        _slice = m.b.b2[:, 'a'].component('vn')[:, 'c', 3, :, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[*,a].vn[*,c,3,*,*]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[1, 'a'].component('vn')[:, 'c', 3, :, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[1,a].vn[*,c,3,*,*]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[...].component('vn')[:, 'c', 3, :, :]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[**].vn[*,c,3,*,*]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

        _slice = m.b.b2[:, 'a'].component('vn')[...]
        cuid = ComponentUID(_slice)
        cuid_str = ComponentUID('b.b2[*,a].vn[**]')
        self.assertEqual(cuid, cuid_str)
        self.assertEqual(str(cuid), str(cuid_str))
        self.assertListSameComponents(m, cuid, cuid_str)

    def test_cuid_from_slice_errors(self):
        # two getitem
        m = self._slice_model()
        m.b.comp = Reference(m.b.b1[:].v1)
        _slice = m.b[:].comp[1][1.1]
        with self.assertRaisesRegex(ValueError, r'.*Two `get_item` calls.*'):
            cuid = ComponentUID(_slice)

        _slice = IndexedComponent_slice(
            m.b[:].component('v'), (IndexedComponent_slice.del_attribute, ('foo',))
        )
        with self.assertRaisesRegex(
            ValueError,
            "Cannot create a CUID from a slice that "
            "contains `set` or `del` calls: got call %s "
            r"with argument \('foo',\)" % (IndexedComponent_slice.del_attribute,),
        ):
            cuid = ComponentUID(_slice)

    def test_cuid_from_cuid(self):
        def assert_equal(cuid1, cuid2):
            self.assertEqual(cuid1, cuid2)
            self.assertFalse(cuid1 is cuid2)

        cuid_str = ComponentUID("b.var[1]")
        cuid_str_2 = ComponentUID(cuid_str)
        assert_equal(cuid_str, cuid_str_2)

        cuid_comp = ComponentUID(self.m.b[1, 1].c)
        cuid_comp_2 = ComponentUID(cuid_comp)
        assert_equal(cuid_str, cuid_str_2)

        cuid_slice = ComponentUID(self.m.b[1, :].c)
        cuid_slice_2 = ComponentUID(cuid_slice)
        assert_equal(cuid_slice, cuid_slice_2)

        with self.assertRaisesRegex(ValueError, "Context is not allowed"):
            ComponentUID(cuid_comp, context=self.m.b[1, 1])


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc
import pickle

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap
from pyomo.core.kernel.variable import variable, variable_dict, variable_list
from pyomo.core.kernel.constraint import constraint, constraint_dict, constraint_list
from pyomo.core.kernel.objective import objective, objective_dict, objective_list
from pyomo.core.kernel.expression import expression, expression_dict, expression_list
from pyomo.core.kernel.block import block, block_dict, block_list
from pyomo.core.kernel.suffix import suffix


class TestComponentMap(unittest.TestCase):
    _components = [
        (variable(), "v"),
        (variable_dict(), "vdict"),
        (variable_list(), "vlist"),
        (constraint(), "c"),
        (constraint_dict(), "cdict"),
        (constraint_list(), "clist"),
        (objective(), "o"),
        (objective_dict(), "odict"),
        (objective_list(), "olist"),
        (expression(), "e"),
        (expression_dict(), "edict"),
        (expression_list(), "elist"),
        (block(), "b"),
        (block_dict(), "bdict"),
        (block_list(), "blist"),
        (suffix(), "s"),
    ]

    def test_pickle(self):
        c = ComponentMap()
        self.assertEqual(len(c), 0)
        cup = pickle.loads(pickle.dumps(c))
        self.assertIsNot(cup, c)
        self.assertEqual(len(cup), 0)

        v = variable()
        c[v] = 1.0
        self.assertEqual(len(c), 1)
        self.assertEqual(c[v], 1.0)
        cup = pickle.loads(pickle.dumps(c))
        vup = list(cup.keys())[0]
        self.assertIsNot(cup, c)
        self.assertIsNot(vup, v)
        self.assertEqual(len(cup), 1)
        self.assertEqual(cup[vup], 1)
        self.assertEqual(vup.parent, None)

        b = block()
        V = b.V = variable_list()
        b.V.append(v)
        b.c = c
        self.assertEqual(len(c), 1)
        self.assertEqual(c[v], 1.0)
        self.assertIs(v.parent, b.V)
        self.assertIs(V.parent, b)
        self.assertIs(b.parent, None)
        bup = pickle.loads(pickle.dumps(b))
        Vup = bup.V
        vup = Vup[0]
        cup = bup.c
        self.assertIsNot(cup, c)
        self.assertIsNot(vup, v)
        self.assertIsNot(Vup, V)
        self.assertIsNot(bup, b)
        self.assertEqual(len(cup), 1)
        self.assertEqual(cup[vup], 1)
        self.assertIs(vup.parent, Vup)
        self.assertIs(Vup.parent, bup)
        self.assertIs(bup.parent, None)

        self.assertEqual(len(c), 1)
        self.assertEqual(c[v], 1)

    def test_init(self):
        cmap = ComponentMap()
        cmap = ComponentMap(self._components)
        with self.assertRaises(TypeError):
            cmap = ComponentMap(*self._components)

    def test_type(self):
        cmap = ComponentMap()
        self.assertTrue(isinstance(cmap, collections.abc.Mapping))
        self.assertTrue(isinstance(cmap, collections.abc.MutableMapping))
        self.assertTrue(issubclass(type(cmap), collections.abc.Mapping))
        self.assertTrue(issubclass(type(cmap), collections.abc.MutableMapping))

    def test_str(self):
        cmap = ComponentMap()
        self.assertEqual(str(cmap), "ComponentMap({})")
        cmap.update(self._components)
        str(cmap)

    def test_len(self):
        cmap = ComponentMap()
        self.assertEqual(len(cmap), 0)
        cmap.update(self._components)
        self.assertEqual(len(cmap), len(self._components))
        cmap = ComponentMap(self._components)
        self.assertEqual(len(cmap), len(self._components))
        self.assertTrue(len(self._components) > 0)

    def test_getsetdelitem(self):
        cmap = ComponentMap()
        for c, val in self._components:
            self.assertTrue(c not in cmap)
        for c, val in self._components:
            cmap[c] = val
            self.assertEqual(cmap[c], val)
            self.assertEqual(cmap.get(c), val)
            del cmap[c]
            with self.assertRaises(KeyError):
                cmap[c]
            with self.assertRaises(KeyError):
                del cmap[c]
            self.assertEqual(cmap.get(c), None)

    def test_iter(self):
        cmap = ComponentMap()
        self.assertEqual(list(iter(cmap)), [])
        cmap.update(self._components)
        ids_seen = set()
        for c in cmap:
            ids_seen.add(id(c))
        self.assertEqual(ids_seen, set(id(c) for c, val in self._components))

    def test_keys(self):
        cmap = ComponentMap(self._components)
        self.assertEqual(
            sorted(cmap.keys(), key=id),
            sorted(list(c for c, val in self._components), key=id),
        )

    def test_values(self):
        cmap = ComponentMap(self._components)
        self.assertEqual(
            sorted(cmap.values()), sorted(list(val for c, val in self._components))
        )

    def test_items(self):
        cmap = ComponentMap(self._components)
        for x in cmap.items():
            self.assertEqual(type(x), tuple)
            self.assertEqual(len(x), 2)
        self.assertEqual(
            sorted(cmap.items(), key=lambda _x: (id(_x[0]), _x[1])),
            sorted(self._components, key=lambda _x: (id(_x[0]), _x[1])),
        )

    def test_update(self):
        cmap = ComponentMap()
        self.assertEqual(len(cmap), 0)
        cmap.update(self._components)
        self.assertEqual(len(cmap), len(self._components))
        for c, val in self._components:
            self.assertEqual(cmap[c], val)

    def test_clear(self):
        cmap = ComponentMap()
        self.assertEqual(len(cmap), 0)
        cmap.update(self._components)
        self.assertEqual(len(cmap), len(self._components))
        cmap.clear()
        self.assertEqual(len(cmap), 0)

    def test_setdefault(self):
        cmap = ComponentMap()
        for c, _ in self._components:
            with self.assertRaises(KeyError):
                cmap[c]
            self.assertTrue(c not in cmap)
            cmap.setdefault(c, []).append(1)
            self.assertEqual(cmap[c], [1])
            del cmap[c]
            with self.assertRaises(KeyError):
                cmap[c]
            self.assertTrue(c not in cmap)
            cmap[c] = []
            cmap.setdefault(c, []).append(1)
            self.assertEqual(cmap[c], [1])

    def test_eq(self):
        cmap1 = ComponentMap()
        self.assertNotEqual(cmap1, set())
        self.assertFalse(cmap1 == set())
        self.assertNotEqual(cmap1, list())
        self.assertFalse(cmap1 == list())
        self.assertNotEqual(cmap1, tuple())
        self.assertFalse(cmap1 == tuple())
        self.assertEqual(cmap1, dict())
        self.assertTrue(cmap1 == dict())

        cmap1.update(self._components)
        self.assertNotEqual(cmap1, set())
        self.assertFalse(cmap1 == set())
        self.assertNotEqual(cmap1, list())
        self.assertFalse(cmap1 == list())
        self.assertNotEqual(cmap1, tuple())
        self.assertFalse(cmap1 == tuple())
        self.assertNotEqual(cmap1, dict())
        self.assertFalse(cmap1 == dict())

        self.assertTrue(cmap1 == cmap1)
        self.assertEqual(cmap1, cmap1)

        cmap2 = ComponentMap(self._components)
        self.assertTrue(cmap2 == cmap1)
        self.assertFalse(cmap2 != cmap1)
        self.assertEqual(cmap2, cmap1)
        self.assertTrue(cmap1 == cmap2)
        self.assertFalse(cmap1 != cmap2)
        self.assertEqual(cmap1, cmap2)

        del cmap2[self._components[0][0]]
        self.assertFalse(cmap2 == cmap1)
        self.assertTrue(cmap2 != cmap1)
        self.assertNotEqual(cmap2, cmap1)
        self.assertFalse(cmap1 == cmap2)
        self.assertTrue(cmap1 != cmap2)
        self.assertNotEqual(cmap1, cmap2)

        cmap2 = ComponentMap(self._components)
        o = objective()
        cmap1[o] = 10
        cmap2[o] = 10
        self.assertEqual(cmap1, cmap2)
        cmap2[o] = 20
        self.assertNotEqual(cmap1, cmap2)
        cmap2[o] = 10
        self.assertEqual(cmap1, cmap2)
        del cmap2[o]
        self.assertNotEqual(cmap1, cmap2)
        cmap2[objective()] = 10
        self.assertNotEqual(cmap1, cmap2)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import re
import time
import logging
import subprocess

from pyomo.common import Executable
from pyomo.common.enums import maximize, minimize
from pyomo.common.errors import ApplicationError
from pyomo.common.collections import Bunch
from pyomo.common.tempfiles import TempfileManager

from pyomo.core.kernel.block import IBlock
from pyomo.core import Var
from pyomo.opt.base import ProblemFormat, ResultsFormat, OptSolver
from pyomo.opt.base.solvers import _extract_version, SolverFactory
from pyomo.opt.results import (
    SolverResults,
    SolverStatus,
    TerminationCondition,
    SolutionStatus,
    Solution,
)
from pyomo.opt.solver import SystemCallSolver
from pyomo.solvers.mockmip import MockMIP

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('cbc', doc='The CBC LP/MIP solver')
class CBC(OptSolver):
    """The CBC LP/MIP solver"""

    def __new__(cls, *args, **kwds):
        mode = kwds.pop('solver_io', 'lp')

        if mode == 'lp' or mode is None:
            opt = SolverFactory('_cbc_shell', **kwds)
            opt.set_problem_format(ProblemFormat.cpxlp)
            return opt
        # CBC's MPS parser seems too buggy to expose
        # this option
        # if mode == 'mps':
        #     # *NOTE: CBC uses the COIN-OR MPS reader,
        #     #        which ignores any objective sense
        #     #        declared in the OBJSENSE section.
        #     opt = SolverFactory('_cbc_shell', **kwds)
        #     opt.set_problem_format(ProblemFormat.mps)
        #     return opt
        #
        if mode == 'nl':
            # CBC doesn't not accept all asl style command line
            # options (-s in particular, which is required for
            # streaming output of all asl solvers). Therefore we need
            # to send it through the cbc_shell instead of ASL
            opt = SolverFactory('_cbc_shell', **kwds)
            opt.set_problem_format(ProblemFormat.nl)
            return opt
        elif mode == 'os':
            opt = SolverFactory('_ossolver', **kwds)
            opt.set_options('solver=cbc')
            return opt
        else:
            logger.error('Unknown IO type: %s' % mode)
            return


@SolverFactory.register('_cbc_shell', doc='Shell interface to the CBC LP/MIP solver')
class CBCSHELL(SystemCallSolver):
    """Shell interface to the CBC LP/MIP solver"""

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds['type'] = 'cbc'
        super(CBCSHELL, self).__init__(**kwds)

        # NOTE: eventually both of the following attributes should be
        # migrated to a common base class.  is the current solve
        # warm-started? a transient data member to communicate state
        # information across the _presolve, _apply_solver, and
        # _postsolve methods.
        self._warm_start_solve = False
        # related to the above, the temporary name of the SOLN
        # warm-start file (if any).
        self._warm_start_file_name = None

        #
        # Set up valid problem formats and valid results for each
        # problem format
        #
        self._valid_problem_formats = [
            ProblemFormat.cpxlp,
            ProblemFormat.nl,
            # ProblemFormat.mps,
        ]
        self._valid_result_formats = {
            ProblemFormat.cpxlp: [ResultsFormat.soln],
            ProblemFormat.nl: [ResultsFormat.sol],
            # ProblemFormat.mps: [ResultsFormat.soln],
        }

        # Note: Undefined capabilities default to 'None'
        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.integer = True
        # The quadratic capabilities may be true but there is
        # some weirdness in the solution file that this
        # plugin does not handle correctly  (extra variables
        # added that are not in the symbol map?)
        self._capabilities.quadratic_objective = False
        self._capabilities.quadratic_constraint = False
        # These flags are updated by the set_problem_format method
        # as cbc can handle SOS constraints with the NL file format but
        # currently not through the LP file format
        self._capabilities.sos1 = False
        self._capabilities.sos2 = False

        self.set_problem_format(ProblemFormat.cpxlp)

    def set_problem_format(self, format):
        super(CBCSHELL, self).set_problem_format(format)
        if self._problem_format == ProblemFormat.cpxlp:
            self._capabilities.sos1 = False
            self._capabilities.sos2 = False
        else:
            self._capabilities.sos1 = True
            self._capabilities.sos2 = True
        if self._problem_format == ProblemFormat.nl:
            if self._compiled_with_asl():
                _ver = self.version()
                # It is possible that _ver is None if the parse failed
                if not _ver or _ver[:3] < (2, 7, 0):
                    if _ver is None:
                        _ver_str = "<unknown>"
                    else:
                        _ver_str = '.'.join(str(i) for i in _ver)
                    logger.warning(
                        f"found CBC version {_ver_str} < 2.7; "
                        "ASL support disabled (falling back on LP interface)."
                    )
                    logger.warning("Upgrade CBC to activate ASL support in this plugin")
                    # Fall back on LP
                    self.set_problem_format(ProblemFormat.cpxlp)
            else:
                logger.warning(
                    "CBC solver is not compiled with ASL "
                    "interface (falling back on LP interface)."
                )
                # Fall back on LP
                self.set_problem_format(ProblemFormat.cpxlp)

    def _default_results_format(self, prob_format):
        if prob_format == ProblemFormat.nl:
            return ResultsFormat.sol
        return ResultsFormat.soln

    def warm_start_capable(self):
        if self._problem_format != ProblemFormat.cpxlp:
            return False
        _ver = self.version()
        return _ver and _ver >= (2, 8, 0, 0)

    def _write_soln_file(self, instance, filename):
        # Maybe this could be a useful method for any instance.

        if isinstance(instance, IBlock):
            smap = getattr(instance, "._symbol_maps")[self._smap_id]
        else:
            smap = instance.solutions.symbol_map[self._smap_id]
        byObject = smap.byObject

        column_index = 0
        with open(filename, 'w') as solnfile:
            for var in instance.component_data_objects(Var):
                # Cbc only expects integer variables with non-zero
                # values for mipstart.
                if (
                    var.value
                    and (var.is_integer() or var.is_binary())
                    and (id(var) in byObject)
                ):
                    name = byObject[id(var)]
                    solnfile.write('{} {} {}\n'.format(column_index, name, var.value))
                    # Cbc ignores column indexes, so the value does not matter.
                    column_index += 1

    #
    # Write a warm-start file in the SOLN format.
    #
    def _warm_start(self, instance):
        self._write_soln_file(instance, self._warm_start_file_name)

    # over-ride presolve to extract the warm-start keyword, if specified.
    def _presolve(self, *args, **kwds):
        # create a context in the temporary file manager for
        # this plugin - is "pop"ed in the _postsolve method.
        TempfileManager.push()

        # if the first argument is a string (representing a filename),
        # then we don't have an instance => the solver is being applied
        # to a file.
        self._warm_start_solve = kwds.pop('warmstart', False)
        self._warm_start_file_name = kwds.pop('warmstart_file', None)
        user_warmstart = False
        if self._warm_start_file_name is not None:
            user_warmstart = True

        # the input argument can currently be one of two things: an
        # instance or a filename.  if a filename is provided and a
        # warm-start is indicated, we go ahead and create the temporary
        # file - assuming that the user has already, via some external
        # mechanism, invoked warm_start() with a instance to create the
        # warm start file.
        if self._warm_start_solve and isinstance(args[0], str):
            # we assume the user knows what they are doing...
            pass
        elif self._warm_start_solve and (not isinstance(args[0], str)):
            # assign the name of the warm start file *before* calling
            # the base class presolve - the base class method ends up
            # creating the command line, and the warm start file-name is
            # (obviously) needed there.
            if self._warm_start_file_name is None:
                assert not user_warmstart
                self._warm_start_file_name = TempfileManager.create_tempfile(
                    suffix='.cbc.soln'
                )

        # CBC does not cleanly handle windows-style drive names in the
        # MIPSTART file name (though at least 2.10.5).
        #
        # See https://github.com/coin-or/Cbc/issues/32
        # The problematic source is https://github.com/coin-or/Cbc/blob/3dcedb27664ae458990e9d4d50bc11c2c55917a0/src/CbcSolver.cpp#L9445-L9459
        if self._warm_start_file_name is not None:
            _drive, _path = os.path.splitdrive(self._warm_start_file_name)
            if _drive:
                _cwd_drive = os.path.splitdrive(os.getcwd())[0]
                if _cwd_drive.lower() == _drive.lower():
                    self._warm_start_file_name = _path
                else:
                    logger.warning(
                        "warmstart_file points to a file on a drive "
                        "different from the current working directory.  "
                        "CBC is likely to (silently) ignore the warmstart."
                    )

        # let the base class handle any remaining keywords/actions.
        # let the base class handle any remaining keywords/actions.
        super(CBCSHELL, self)._presolve(*args, **kwds)

        # NB: we must let the base class presolve run first so that the
        # symbol_map is actually constructed!

        if (len(args) > 0) and (not isinstance(args[0], str)):
            if len(args) != 1:
                raise ValueError(
                    "CBCplugin _presolve method can only handle a single "
                    "problem instance - %s were supplied" % (len(args),)
                )

            # write the warm-start file - currently only supports MIPs.
            # we only know how to deal with a single problem instance.
            if self._warm_start_solve and (not user_warmstart):
                start_time = time.time()
                self._warm_start(args[0])
                end_time = time.time()
                if self._report_timing is True:
                    print(
                        "Warm start write time=%.2f seconds" % (end_time - start_time)
                    )

    def _default_executable(self):
        executable = Executable("cbc")
        if not executable:
            logger.warning(
                "Could not locate the 'cbc' executable, which is "
                "required for solver %s" % self.name
            )
            self.enable = False
            return None
        return executable.path()

    def _get_version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        results = subprocess.run(
            [self.executable(), "-stop"],
            timeout=5,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        _version = _extract_version(results.stdout)
        if _version is None:
            return _extract_version('')
        return _version

    def _compiled_with_asl(self):
        results = subprocess.run(
            [self.executable(), "dummy", "-AMPL", "-stop"],
            timeout=5,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        return 'No match for AMPL'.lower() not in results.stdout.lower()

    def create_command_line(self, executable, problem_files):
        #
        # Define the log file
        #
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix=".cbc.log")

        #
        # Define the solution file
        #
        # the prefix of the problem filename is required because CBC has a specific
        # and automatic convention for generating the output solution filename.
        # the extracted prefix is the same name as the input filename, e.g., minus
        # the ".lp" extension.
        problem_filename_prefix = problem_files[0]
        if '.' in problem_filename_prefix:
            tmp = problem_filename_prefix.split('.')
            if len(tmp) > 2:
                problem_filename_prefix = '.'.join(tmp[:-1])
            else:
                problem_filename_prefix = tmp[0]
        if self._results_format is ResultsFormat.sol:
            self._soln_file = problem_filename_prefix + ".sol"
        else:
            self._soln_file = problem_filename_prefix + ".soln"

        #
        # Define the results file (if the sol external parser is used)
        #
        # results in CBC are split across the log file (solver statistics) and
        # the solution file (solutions!)
        if self._results_format is ResultsFormat.sol:
            self._results_file = self._soln_file

        def _check_and_escape_options(options):
            for key, val in self.options.items():
                tmp_k = str(key)
                _bad = ' ' in tmp_k

                tmp_v = str(val)
                if ' ' in tmp_v:
                    if '"' in tmp_v:
                        if "'" in tmp_v:
                            _bad = True
                        else:
                            tmp_v = "'" + tmp_v + "'"
                    else:
                        tmp_v = '"' + tmp_v + '"'

                if _bad:
                    raise ValueError(
                        "Unable to properly escape solver option:"
                        "\n\t%s=%s" % (key, val)
                    )
                yield (tmp_k, tmp_v)

        #
        # Define command line
        #
        cmd = [executable]
        if self._timer:
            cmd.insert(0, self._timer)
        if self._problem_format == ProblemFormat.nl:
            cmd.append(problem_files[0])
            cmd.append('-AMPL')

            if self._timelimit is not None and self._timelimit > 0.0:
                cmd.extend(['-sec', str(self._timelimit)])
                cmd.extend(['-timeMode', "elapsed"])
            if "debug" in self.options:
                cmd.extend(["-log", "5"])
            for key, val in _check_and_escape_options(self.options):
                if key == 'solver':
                    continue
                cmd.append(key + "=" + val)
            os.environ['cbc_options'] = "printingOptions=all"
            # cmd.extend(["-printingOptions=all",
            # "-stat"])
        else:
            if self._timelimit is not None and self._timelimit > 0.0:
                cmd.extend(['-sec', str(self._timelimit)])
                cmd.extend(['-timeMode', "elapsed"])
            if "debug" in self.options:
                cmd.extend(["-log", "5"])
            # these must go after options that take a value
            action_options = []
            for key, val in _check_and_escape_options(self.options):
                if val.strip() != '':
                    cmd.extend(['-' + key, val])
                else:
                    action_options.append('-' + key)
            cmd.extend(["-printingOptions", "all", "-import", problem_files[0]])
            cmd.extend(action_options)
            if self._warm_start_solve:
                cmd.extend(["-mipstart", self._warm_start_file_name])
            cmd.extend(["-stat=1", "-solve", "-solu", self._soln_file])

        return Bunch(cmd=cmd, log_file=self._log_file, env=None)

    def process_logfile(self):
        """
        Process logfile
        """

        results = SolverResults()

        # The logfile output for cbc when using nl files
        # provides no information worth parsing here
        if self._problem_format is ProblemFormat.nl:
            return results

        #
        # Initial values
        #
        soln = Solution()

        #
        # Process logfile
        #
        OUTPUT = open(self._log_file)
        output = "".join(OUTPUT.readlines())
        OUTPUT.close()
        #
        # Parse logfile lines
        #
        results.problem.sense = minimize
        results.problem.name = None
        optim_value = float('inf')
        lower_bound = None
        upper_bound = None
        gap = None
        nodes = None
        # See https://www.coin-or.org/Cbc/cbcuserguide.html#messages
        for line in output.split("\n"):
            tokens = tuple(re.split('[ \t]+', line.strip()))
            n_tokens = len(tokens)
            if n_tokens > 1:
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L3769
                if n_tokens > 4 and tokens[:4] == (
                    'Continuous',
                    'objective',
                    'value',
                    'is',
                ):
                    lower_bound = _float(tokens[4])
                # Search completed - best objective %g, took %d iterations and %d nodes
                elif (
                    n_tokens > 12
                    and tokens[1:3] == ('Search', 'completed')
                    and tokens[4:6] == ('best', 'objective')
                    and tokens[9] == 'iterations'
                    and tokens[12] == 'nodes'
                ):
                    optim_value = _float(tokens[6][:-1])
                    results.solver.statistics.black_box.number_of_iterations = int(
                        tokens[8]
                    )
                    nodes = int(tokens[11])
                elif tokens[1] == 'Exiting' and n_tokens > 4:
                    if tokens[2:4] == ('on', 'maximum'):
                        results.solver.termination_condition = {
                            'nodes': TerminationCondition.maxEvaluations,
                            'time': TerminationCondition.maxTimeLimit,
                            'solutions': TerminationCondition.other,
                            'iterations': TerminationCondition.maxIterations,
                        }.get(tokens[4], TerminationCondition.other)
                    # elif tokens[2:5] == ('as', 'integer', 'gap'):
                    #     # We might want to handle this case
                # Integer solution of %g found...
                elif n_tokens >= 4 and tokens[1:4] == ('Integer', 'solution', 'of'):
                    optim_value = _float(tokens[4])
                    try:
                        results.solver.statistics.black_box.number_of_iterations = int(
                            tokens[tokens.index('iterations') - 1]
                        )
                        nodes = int(tokens[tokens.index('nodes') - 1])
                    except ValueError:
                        pass
                # Partial search - best objective %g (best possible %g), took %d iterations and %d nodes
                elif (
                    n_tokens > 15
                    and tokens[1:3] == ('Partial', 'search')
                    and tokens[4:6] == ('best', 'objective')
                    and tokens[7:9] == ('(best', 'possible')
                    and tokens[12] == 'iterations'
                    and tokens[15] == 'nodes'
                ):
                    optim_value = _float(tokens[6])
                    lower_bound = _float(tokens[9][:-2])
                    results.solver.statistics.black_box.number_of_iterations = int(
                        tokens[11]
                    )
                    nodes = int(tokens[14])
                elif (
                    n_tokens > 12
                    and tokens[1] == 'After'
                    and tokens[3] == 'nodes,'
                    and tokens[8:10] == ('best', 'solution,')
                    and tokens[10:12] == ('best', 'possible')
                ):
                    nodes = int(tokens[2])
                    optim_value = _float(tokens[7])
                    lower_bound = _float(tokens[12])
                elif (
                    tokens[0] == "Current"
                    and n_tokens == 10
                    and tokens[1] == "default"
                    and tokens[2] == "(if"
                    and results.problem.name is None
                ):
                    results.problem.name = tokens[-1]
                    if '.' in results.problem.name:
                        parts = results.problem.name.split('.')
                        if len(parts) > 2:
                            results.problem.name = '.'.join(parts[:-1])
                        else:
                            results.problem.name = results.problem.name.split('.')[0]
                    if '/' in results.problem.name:
                        results.problem.name = results.problem.name.split('/')[-1]
                    if '\\' in results.problem.name:
                        results.problem.name = results.problem.name.split('\\')[-1]
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L10840
                elif tokens[0] == 'Presolve':
                    if n_tokens > 9 and tokens[3] == 'rows,' and tokens[6] == 'columns':
                        results.problem.number_of_variables = int(tokens[4]) - int(
                            tokens[5][1:-1]
                        )
                        results.problem.number_of_constraints = int(tokens[1]) - int(
                            tokens[2][1:-1]
                        )
                        results.problem.number_of_objectives = 1
                    elif n_tokens > 6 and tokens[6] == 'infeasible':
                        soln.status = SolutionStatus.infeasible
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L11105
                elif (
                    n_tokens > 11
                    and tokens[:2] == ('Problem', 'has')
                    and tokens[3] == 'rows,'
                    and tokens[5] == 'columns'
                    and tokens[7:9] == ('with', 'objective)')
                ):
                    results.problem.number_of_variables = int(tokens[4])
                    results.problem.number_of_constraints = int(tokens[2])
                    results.problem.number_of_nonzeros = int(tokens[6][1:])
                    results.problem.number_of_objectives = 1
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L10814
                elif (
                    n_tokens > 8
                    and tokens[:3] == ('Original', 'problem', 'has')
                    and tokens[4] == 'integers'
                    and tokens[6:9] == ('of', 'which', 'binary)')
                ):
                    results.problem.number_of_integer_variables = int(tokens[3])
                    results.problem.number_of_binary_variables = int(tokens[5][1:])
                elif n_tokens == 5 and tokens[3] == "NAME":
                    results.problem.name = tokens[4]
                elif (
                    'CoinLpIO::readLp(): Maximization problem reformulated as minimization'
                    in ' '.join(tokens)
                ):
                    results.problem.sense = maximize
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L3047
                elif n_tokens > 3 and tokens[:2] == ('Result', '-'):
                    if tokens[2:4] in [('Run', 'abandoned'), ('User', 'ctrl-c')]:
                        results.solver.termination_condition = (
                            TerminationCondition.userInterrupt
                        )
                    if n_tokens > 4:
                        if tokens[2:5] == ('Optimal', 'solution', 'found'):
                            # parser for log file generetated with discrete variable
                            soln.status = SolutionStatus.optimal
                            # if n_tokens > 7 and tokens[5:8] == ('(within', 'gap', 'tolerance)'):
                            #     # We might want to handle this case
                        elif tokens[2:5] in [
                            ('Linear', 'relaxation', 'infeasible'),
                            ('Problem', 'proven', 'infeasible'),
                        ]:
                            soln.status = SolutionStatus.infeasible
                        elif tokens[2:5] == ('Linear', 'relaxation', 'unbounded'):
                            soln.status = SolutionStatus.unbounded
                        elif (
                            n_tokens > 5
                            and tokens[2:4] == ('Stopped', 'on')
                            and tokens[5] == 'limit'
                        ):
                            results.solver.termination_condition = {
                                'node': TerminationCondition.maxEvaluations,
                                'time': TerminationCondition.maxTimeLimit,
                                'solution': TerminationCondition.other,
                                'iterations': TerminationCondition.maxIterations,
                            }.get(tokens[4], TerminationCondition.other)
                    # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L12318
                    elif n_tokens > 3 and tokens[2] == "Finished":
                        soln.status = SolutionStatus.optimal
                        optim_value = _float(tokens[4])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7904
                elif n_tokens >= 3 and tokens[:2] == ('Objective', 'value:'):
                    # parser for log file generetated with discrete variable
                    optim_value = _float(tokens[2])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7904
                elif n_tokens >= 4 and tokens[:4] == (
                    'No',
                    'feasible',
                    'solution',
                    'found',
                ):
                    soln.status = SolutionStatus.infeasible
                elif n_tokens > 2 and tokens[:2] == ('Lower', 'bound:'):
                    if (
                        lower_bound is None
                    ):  # Only use if not already found since this is to less decimal places
                        results.problem.lower_bound = _float(tokens[2])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7918
                elif tokens[0] == 'Gap:':
                    # This is relative and only to 2 decimal places - could calculate explicitly using lower bound
                    gap = _float(tokens[1])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7923
                elif n_tokens > 2 and tokens[:2] == ('Enumerated', 'nodes:'):
                    nodes = int(tokens[2])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7926
                elif n_tokens > 2 and tokens[:2] == ('Total', 'iterations:'):
                    results.solver.statistics.black_box.number_of_iterations = int(
                        tokens[2]
                    )
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7930
                elif n_tokens > 3 and tokens[:3] == ('Time', '(CPU', 'seconds):'):
                    results.solver.system_time = _float(tokens[3])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L7933
                elif n_tokens > 3 and tokens[:3] == ('Time', '(Wallclock', 'Seconds):'):
                    results.solver.wallclock_time = _float(tokens[3])
                # https://github.com/coin-or/Cbc/blob/cb6bf98/Cbc/src/CbcSolver.cpp#L10477
                elif n_tokens > 4 and tokens[:4] == (
                    'Total',
                    'time',
                    '(CPU',
                    'seconds):',
                ):
                    results.solver.system_time = _float(tokens[4])
                    if n_tokens > 7 and tokens[5:7] == ('(Wallclock', 'seconds):'):
                        results.solver.wallclock_time = _float(tokens[7])
                elif tokens[0] == "Optimal":
                    if n_tokens > 4 and tokens[2] == "objective" and tokens[4] != "and":
                        # parser for log file generetated without discrete variable
                        # see pull request #339: last check avoids lines like "Optimal - objective gap and
                        # complementarity gap both smallish and small steps"
                        soln.status = SolutionStatus.optimal
                        optim_value = _float(tokens[4])
                    elif (
                        n_tokens > 5
                        and tokens[1] == 'objective'
                        and tokens[5] == 'iterations'
                    ):
                        soln.status = SolutionStatus.optimal
                        optim_value = _float(tokens[2])
                        results.solver.statistics.black_box.number_of_iterations = int(
                            tokens[4]
                        )
                elif tokens[0] == "sys" and n_tokens == 2:
                    results.solver.system_time = _float(tokens[1])
                elif tokens[0] == "user" and n_tokens == 2:
                    results.solver.user_time = _float(tokens[1])
                elif (
                    n_tokens == 10
                    and "Presolve" in tokens
                    and "iterations" in tokens
                    and tokens[0] == "Optimal"
                    and "objective" == tokens[1]
                ):
                    soln.status = SolutionStatus.optimal
                    optim_value = _float(tokens[2])
                results.solver.user_time = -1.0  # Why is this set to -1?

        if results.problem.name is None:
            results.problem.name = 'unknown'

        if soln.status is SolutionStatus.optimal:
            results.solver.termination_message = (
                "Model was solved to optimality (subject to tolerances), and an "
                "optimal solution is available."
            )
            results.solver.termination_condition = TerminationCondition.optimal
            results.solver.status = SolverStatus.ok
            if gap is None:
                lower_bound = optim_value
                gap = 0.0
        elif soln.status == SolutionStatus.infeasible:
            results.solver.termination_message = "Model was proven to be infeasible."
            results.solver.termination_condition = TerminationCondition.infeasible
            results.solver.status = SolverStatus.warning
        elif soln.status == SolutionStatus.unbounded:
            results.solver.termination_message = "Model was proven to be unbounded."
            results.solver.termination_condition = TerminationCondition.unbounded
            results.solver.status = SolverStatus.warning
        elif results.solver.termination_condition in [
            TerminationCondition.maxTimeLimit,
            TerminationCondition.maxEvaluations,
            TerminationCondition.other,
            TerminationCondition.maxIterations,
        ]:
            results.solver.status = SolverStatus.aborted
            soln.status = SolutionStatus.stoppedByLimit
            if (
                results.solver.termination_condition
                == TerminationCondition.maxTimeLimit
            ):
                results.solver.termination_message = (
                    "Optimization terminated because the time expended "
                    "exceeded the value specified in the seconds "
                    "parameter."
                )
            elif (
                results.solver.termination_condition
                == TerminationCondition.maxEvaluations
            ):
                results.solver.termination_message = (
                    "Optimization terminated because the total number of branch-and-cut nodes explored "
                    "exceeded the value specified in the maxNodes parameter"
                )
            elif results.solver.termination_condition == TerminationCondition.other:
                results.solver.termination_message = (
                    "Optimization terminated because the number of "
                    "solutions found reached the value specified in the "
                    "maxSolutions parameter."
                )
            elif (
                results.solver.termination_condition
                == TerminationCondition.maxIterations
            ):
                results.solver.termination_message = (
                    "Optimization terminated because the total number of simplex "
                    "iterations performed exceeded the value specified in the "
                    "maxIterations parameter."
                )
        soln.gap = gap
        if results.problem.sense == minimize:
            upper_bound = optim_value
        elif results.problem.sense == maximize:
            _ver = self.version()
            if _ver and _ver[:3] < (2, 10, 2):
                optim_value *= -1
                upper_bound = None if lower_bound is None else -lower_bound
            else:
                upper_bound = None if lower_bound is None else lower_bound
            lower_bound = optim_value
        soln.objective['__default_objective__'] = {'Value': optim_value}
        results.problem.lower_bound = lower_bound
        results.problem.upper_bound = upper_bound

        results.solver.statistics.branch_and_bound.number_of_bounded_subproblems = nodes
        results.solver.statistics.branch_and_bound.number_of_created_subproblems = nodes

        if soln.status in [
            SolutionStatus.optimal,
            SolutionStatus.stoppedByLimit,
            SolutionStatus.unknown,
            SolutionStatus.other,
        ]:
            results.solution.insert(soln)

        return results

    def process_soln_file(self, results):
        # the only suffixes that we extract from CBC are
        # constraint duals and variable reduced-costs. scan
        # through the solver suffix list and throw an
        # exception if the user has specified any others.
        extract_duals = False
        extract_reduced_costs = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "rc"):
                extract_reduced_costs = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***CBC solver plugin cannot extract solution suffix=" + suffix
                )

        # if dealing with SOL format files, we've already read
        # this via the base class reader functionality.
        if self._results_format is ResultsFormat.sol:
            return

        # otherwise, go with the native CBC solution format.
        if len(results.solution) > 0:
            solution = results.solution(0)
        else:
            solution = Solution()

        results.problem.number_of_objectives = 1

        processing_constraints = (
            None  # None means header, True means constraints, False means variables.
        )
        header_processed = False
        optim_value = None

        try:
            INPUT = open(self._soln_file, "r")
        except IOError:
            INPUT = []

        _ver = self.version()
        invert_objective_sense = results.problem.sense == maximize and (
            _ver and _ver[:3] < (2, 10, 2)
        )

        for line in INPUT:
            tokens = tuple(re.split('[ \t]+', line.strip()))
            n_tokens = len(tokens)
            #
            # These are the only header entries CBC will generate
            # (identified via browsing CbcSolver.cpp).  See
            #     https://github.com/coin-or/Cbc/tree/master/src/CbcSolver.cpp
            # Search for "(no integer solution - continuous used)"
            # (L10796 as of cb855c7)
            #
            # Note that since this possibly also supports old CBC
            # versions, we shall not be removing any functionality, even
            # if it is not seen in the current revision
            #
            if not header_processed:
                if tokens[0] == 'Optimal':
                    results.solver.termination_condition = TerminationCondition.optimal
                    results.solver.status = SolverStatus.ok
                    results.solver.termination_message = (
                        "Model was solved to optimality (subject to tolerances), "
                        "and an optimal solution is available."
                    )
                    solution.status = SolutionStatus.optimal
                    optim_value = _float(tokens[-1])
                elif tokens[0] in ('Infeasible', 'PrimalInfeasible') or (
                    n_tokens > 1 and tokens[0:2] == ('Integer', 'infeasible')
                ):
                    results.solver.termination_message = (
                        "Model was proven to be infeasible."
                    )
                    results.solver.termination_condition = (
                        TerminationCondition.infeasible
                    )
                    results.solver.status = SolverStatus.warning
                    solution.status = SolutionStatus.infeasible
                    INPUT.close()
                    return
                elif (
                    tokens[0] == 'Unbounded'
                    or (
                        n_tokens > 2
                        and tokens[0] == 'Problem'
                        and tokens[2] == 'unbounded'
                    )
                    or (n_tokens > 1 and tokens[0:2] == ('Dual', 'infeasible'))
                ):
                    results.solver.termination_message = (
                        "Model was proven to be unbounded."
                    )
                    results.solver.termination_condition = (
                        TerminationCondition.unbounded
                    )
                    results.solver.status = SolverStatus.warning
                    solution.status = SolutionStatus.unbounded
                    INPUT.close()
                    return
                elif n_tokens > 2 and tokens[0:2] == ('Stopped', 'on'):
                    optim_value = _float(tokens[-1])
                    solution.gap = None
                    results.solver.status = SolverStatus.aborted
                    solution.status = SolutionStatus.stoppedByLimit
                    if tokens[2] == 'time':
                        results.solver.termination_message = (
                            "Optimization terminated because the time expended "
                            "exceeded the value specified in the seconds "
                            "parameter."
                        )
                        results.solver.termination_condition = (
                            TerminationCondition.maxTimeLimit
                        )
                    elif tokens[2] == 'iterations':
                        # Only add extra info if not already obtained from logs (which give a better description)
                        if results.solver.termination_condition not in [
                            TerminationCondition.maxEvaluations,
                            TerminationCondition.other,
                            TerminationCondition.maxIterations,
                        ]:
                            results.solver.termination_message = (
                                "Optimization terminated because a limit was hit"
                            )
                            results.solver.termination_condition = (
                                TerminationCondition.maxIterations
                            )
                    elif tokens[2] == 'difficulties':
                        results.solver.termination_condition = (
                            TerminationCondition.solverFailure
                        )
                        results.solver.status = SolverStatus.error
                        solution.status = SolutionStatus.error
                    elif tokens[2] == 'ctrl-c':
                        results.solver.termination_message = (
                            "Optimization was terminated by the user."
                        )
                        results.solver.termination_condition = (
                            TerminationCondition.userInterrupt
                        )
                        solution.status = SolutionStatus.unknown
                    else:
                        results.solver.termination_condition = (
                            TerminationCondition.unknown
                        )
                        results.solver.status = SolverStatus.unknown
                        solution.status = SolutionStatus.unknown
                        results.solver.termination_message = ' '.join(tokens)
                        print(
                            '***WARNING: CBC plugin currently not processing solution status=Stopped correctly. Full '
                            'status line is: {}'.format(line.strip())
                        )
                    if n_tokens > 8 and tokens[3:9] == (
                        '(no',
                        'integer',
                        'solution',
                        '-',
                        'continuous',
                        'used)',
                    ):
                        results.solver.termination_message = (
                            "Optimization terminated because a limit was hit, "
                            "however it had not found an integer solution yet."
                        )
                        results.solver.termination_condition = (
                            TerminationCondition.intermediateNonInteger
                        )
                        solution.status = SolutionStatus.other
                else:
                    results.solver.termination_condition = TerminationCondition.unknown
                    results.solver.status = SolverStatus.unknown
                    solution.status = SolutionStatus.unknown
                    results.solver.termination_message = ' '.join(tokens)
                    print(
                        '***WARNING: CBC plugin currently not processing solution status={} correctly. Full status '
                        'line is: {}'.format(tokens[0], line.strip())
                    )

            # most of the first tokens should be integers
            # if it's not an integer, only then check the list of results
            try:
                row_number = int(tokens[0])
                if row_number == 0:  # indicates section start.
                    if processing_constraints is None:
                        processing_constraints = True
                    elif processing_constraints is True:
                        processing_constraints = False
                    else:
                        raise RuntimeError(
                            "CBC plugin encountered unexpected line=("
                            + line.strip()
                            + ") in solution file="
                            + self._soln_file
                            + "; constraint and variable sections already processed!"
                        )
            except ValueError:
                if tokens[0] in (
                    "Optimal",
                    "Infeasible",
                    "Unbounded",
                    "Stopped",
                    "Integer",
                    "Status",
                ):
                    if optim_value is not None:
                        if invert_objective_sense:
                            optim_value *= -1
                        solution.objective['__default_objective__'] = {
                            'Value': optim_value
                        }
                    header_processed = True

            if (processing_constraints is True) and (extract_duals is True):
                if n_tokens == 4:
                    pass
                elif (n_tokens == 5) and tokens[0] == "**":
                    tokens = tokens[1:]
                else:
                    raise RuntimeError(
                        "Unexpected line format encountered in CBC solution file - line="
                        + line
                    )

                constraint = tokens[1]
                constraint_ax = _float(
                    tokens[2]
                )  # CBC reports the constraint row times the solution vector - not the slack.
                constraint_dual = _float(tokens[3])
                if invert_objective_sense:
                    constraint_dual *= -1
                if constraint[:2] == 'c_':
                    solution.constraint[constraint] = {"Dual": constraint_dual}
                elif constraint[:2] == 'r_':
                    # For the range constraints, supply only the dual with the largest
                    # magnitude (at least one should always be numerically zero)
                    existing_constraint_dual_dict = solution.constraint.get(
                        'r_l_' + constraint[4:], None
                    )
                    if existing_constraint_dual_dict:
                        # if a constraint dual is already saved, then update it if its
                        # magnitude is larger than existing; this avoids checking vs
                        # zero (avoiding problems with solver tolerances)
                        existing_constraint_dual = existing_constraint_dual_dict["Dual"]
                        if abs(constraint_dual) > abs(existing_constraint_dual):
                            solution.constraint['r_l_' + constraint[4:]] = {
                                "Dual": constraint_dual
                            }
                    else:
                        # if no constraint with that name yet, just save it in the solution constraint dictionary
                        solution.constraint['r_l_' + constraint[4:]] = {
                            "Dual": constraint_dual
                        }

            elif processing_constraints is False:
                if n_tokens == 4:
                    pass
                elif (n_tokens == 5) and tokens[0] == "**":
                    tokens = tokens[1:]
                else:
                    raise RuntimeError(
                        "Unexpected line format encountered "
                        "in CBC solution file - line=" + line
                    )

                variable_name = tokens[1]
                variable_value = _float(tokens[2])
                variable = solution.variable[variable_name] = {"Value": variable_value}
                if extract_reduced_costs is True:
                    variable_reduced_cost = _float(tokens[3])  # currently ignored.
                    if invert_objective_sense:
                        variable_reduced_cost *= -1
                    variable["Rc"] = variable_reduced_cost

            elif header_processed is True:
                pass

            else:
                raise RuntimeError(
                    "CBC plugin encountered unexpected "
                    "line=("
                    + line.strip()
                    + ") in solution file="
                    + self._soln_file
                    + "; expecting header, but "
                    "found data!"
                )

        if not type(INPUT) is list:
            INPUT.close()

        if len(results.solution) == 0 and solution.status in [
            SolutionStatus.optimal,
            SolutionStatus.stoppedByLimit,
            SolutionStatus.unknown,
            SolutionStatus.other,
        ]:
            results.solution.insert(solution)

    def _postsolve(self):
        # let the base class deal with returning results.
        results = super(CBCSHELL, self)._postsolve()

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin. does not
        # include, for example, the execution script. but does include
        # the warm-start file.
        TempfileManager.pop(remove=not self._keepfiles)

        return results


def _float(x):
    return float(x) if x != '1.#J' else float('inf')


@SolverFactory.register('_mock_cbc')
class MockCBC(CBCSHELL, MockMIP):
    """A Mock CBC solver used for testing"""

    def __init__(self, **kwds):
        try:
            CBCSHELL.__init__(self, **kwds)
        except ApplicationError:  # pragma:nocover
            pass  # pragma:nocover
        MockMIP.__init__(self, "cbc")

    def available(self, exception_flag=True):
        return CBCSHELL.available(self, exception_flag)

    def create_command_line(self, executable, problem_files):
        command = CBCSHELL.create_command_line(self, executable, problem_files)
        MockMIP.create_command_line(self, executable, problem_files)
        return command

    def executable(self):
        return MockMIP.executable(self)

    def _execute_command(self, cmd):
        return MockMIP._execute_command(self, cmd)

    def _convert_problem(self, args, pformat, valid_pformats):
        if pformat in [ProblemFormat.mps, ProblemFormat.cpxlp, ProblemFormat.nl]:
            return (args, pformat, None)
        else:
            return (args, ProblemFormat.mps, None)

    def _get_version(self):
        return 2, 9, 9
